from multiplex import keys, ansi


class HelpViewState:
    def __init__(self, viewer):
        self.viewer = viewer
        self.show = False
        self.current_line = 0
        self.max_current_line = (
            2  # box lines
            + len(keys.descriptions) * 2
            - 1  # last description new line
            + sum(len(mode_desc) for mode_desc in keys.descriptions.values())
        )

    def toggle(self):
        self.show = not self.show
        self.current_line = 0
        return ansi.FULL_REFRESH

    def move_line_up(self):
        self.current_line = max(0, self.current_line - 1)
        return True

    def move_line_down(self):
        max_line = self.max_current_line
        if self.viewer.lines:
            max_line -= self.viewer.lines
        self.current_line = max(0, min(max_line, self.current_line + 1))
        return True



from dataclasses import dataclass
from typing import Union


@dataclass
class Process:
    cmd: Union[str, list, tuple]

    @property
    def title(self):
        return self.cmd if isinstance(self.cmd, str) else " ".join(self.cmd)



import pytest

from multiplex.keys import generic_key_to_name, seq_to_name, HOME, is_multi, key_to_seq, bind


def test_generic_key_to_name_printable():
    assert generic_key_to_name(ord("k")) == "k"


def test_generic_key_to_name_non_printable():
    assert generic_key_to_name(20) == "[20]"


def test_seq_to_name_list_input():
    assert seq_to_name([ord("k")]) == "k"


def test_seq_to_name_ctrl_name():
    assert seq_to_name((11,)) == "^K"


def test_seq_to_name_predefined():
    assert seq_to_name(HOME) == "Home"


def test_seq_to_name_single_char():
    assert seq_to_name((ord("k"),)) == "k"


def test_seq_to_name_fallback():
    assert seq_to_name((ord("k"), 20)) == "k[20]"


def test_seq_to_name_no_fallback():
    assert seq_to_name((ord("k"), 20), fallback=False) is None


def test_is_multi():
    assert not is_multi((11,))
    assert not is_multi((11, 12))
    assert not is_multi([11, 12])
    assert is_multi([[11], [12]])
    assert is_multi([(11,), (12,)])
    assert is_multi(["one", "two"])


def test_key_to_seq_str_input():
    assert key_to_seq("abc") == (ord("a"), ord("b"), ord("c"))


def test_key_to_seq_multi_input():
    assert key_to_seq(["gg", (11,)]) == (ord("g"), ord("g"), 11)


def test_key_to_seq_standard_input():
    assert key_to_seq([1, 2, 3]) == (1, 2, 3)
    assert key_to_seq((1, 2, 3)) == (1, 2, 3)


def test_key_to_seq_invalid_input():
    with pytest.raises(RuntimeError):
        key_to_seq(True)


def test_bind_basic():
    mode = "a"
    bindings = {mode: {}}
    descriptions = {mode: {}}

    @bind(mode, "a", "b", description="c", custom_bindings=bindings, custom_descriptions=descriptions)
    def fn1():
        pass

    assert bindings[mode][(ord("a"),)] is fn1
    assert bindings[mode][(ord("b"),)] is fn1
    assert descriptions[mode][("a", "b")] == "c"


def test_bind_description_fallback():
    mode = "a"
    bindings = {mode: {}}
    descriptions = {mode: {}}

    @bind(mode, "a", custom_bindings=bindings, custom_descriptions=descriptions)
    def fn1():
        pass

    assert descriptions[mode][("a",)] == "fn1"


def test_bind_multi_description():
    mode = "a"
    bindings = {mode: {}}
    descriptions = {mode: {}}

    @bind(mode, ["a", "b"], custom_bindings=bindings, custom_descriptions=descriptions)
    def fn1():
        pass

    assert descriptions[mode][("ab",)] == "fn1"


def test_bind_seq_to_name_description():
    mode = "a"
    bindings = {mode: {}}
    descriptions = {mode: {}}

    @bind(mode, ["a", (27,)], custom_bindings=bindings, custom_descriptions=descriptions)
    def fn1():
        pass

    assert descriptions[mode][("aEsc",)] == "fn1"



from multiplex.keys import *

from .exceptions import EndViewer

from .ansi import FULL_REFRESH


# toggles
@bind(GLOBAL, "s", description="Toggle auto-scoll/manual-scroll for currently focused box")
def toggle_auto_scroll(viewer):
    return viewer.focused.toggle_auto_scroll()


@bind(GLOBAL, "i", description="Activate input mode for currently focused box")
def activate_input_mode(viewer):
    if viewer.focused.state.stream_done:
        return
    return viewer.focused.activate_input_mode()


@bind(GLOBAL, "w", description="Toggle wrap/unwrap for currently focused box")
def toggle_wrap(viewer):
    return viewer.focused.toggle_wrap()


@bind(GLOBAL, "W", description="Toggle wrap/unwrap for all boxes")
def toggle_wrap_all(viewer):
    if all(box.state.wrap for box in viewer.boxes):
        new_value = False
    elif not any(box.state.wrap for box in viewer.boxes):
        new_value = True
    else:
        new_value = not viewer.wrapped_all
    for box in viewer.boxes:
        box.toggle_wrap(new_value)
    viewer.wrapped_all = new_value
    viewer.verify_focused_box_in_view()
    return FULL_REFRESH


@bind(GLOBAL, "c", description="Toggle expand/collapse for currently focused box")
def toggle_collapse(viewer):
    return viewer.focused.toggle_collapse()


@bind(GLOBAL, "C", description="Toggle expand/collapse for all boxes")
def toggle_collapse_all(viewer):
    if all(box.state.collapsed for box in viewer.boxes):
        new_value = False
    elif not any(box.state.collapsed for box in viewer.boxes):
        new_value = True
    else:
        new_value = not viewer.collaped_all
    for box in viewer.boxes:
        box.toggle_collapse(new_value)
    viewer.collaped_all = new_value
    viewer.verify_focused_box_in_view()
    return FULL_REFRESH


@bind(GLOBAL, "m", description="Toggle maxmize")
def toggle_maximize(viewer):
    viewer.maximized = not viewer.maximized
    return FULL_REFRESH


@bind(GLOBAL, "=", description="Strip empty lines from boxes that finished processing")
def strip_empty_lines(viewer):
    for box in viewer.boxes:
        box.strip_empty_lines()
    all_up(viewer)
    return FULL_REFRESH


@bind(GLOBAL, "O", description="Dump boxes to output_dir (default: $PWD)")
async def save(viewer):
    await viewer.export.save()


@bind(GLOBAL, "?", description="Show/hide this help screen")
def toggle_help(viewer):
    return viewer.help.toggle()


# quit
@bind(GLOBAL, "q", description="Quit")
def end(*_):
    raise EndViewer


# global inside box movement
@bind(GLOBAL, "l", RIGHT, description="Move 1 char right")
def move_right(viewer):
    return viewer.focused.move_right()


@bind(GLOBAL, "h", LEFT, description="Move 1 char left")
def move_left(viewer):
    return viewer.focused.move_left()


@bind(GLOBAL, ")", description="Move 1/2 screen to the right")
def move_half_screen_right(viewer):
    return viewer.focused.move_half_screen_right()


@bind(GLOBAL, "(", description="Move 1/2 screen to the left")
def move_half_screen_left(viewer):
    return viewer.focused.move_half_screen_left()


@bind(GLOBAL, "$", description="Move all the way to the right")
def move_right_until_end(viewer):
    return viewer.focused.move_right_until_end()


@bind(GLOBAL, "_", "0", description="Move all the way to the left")
def move_left_until_start(viewer):
    return viewer.focused.move_left_until_start()


# main mode
@bind(NORMAL, "j", DOWN, description="Focus next box")
def next_box(viewer):
    viewer.current_focused_box = min(viewer.current_focused_box + 1, viewer.num_boxes - 1)
    viewer.verify_focused_box_in_view()


@bind(NORMAL, "k", UP, description="Focus previous box")
def previous_box(viewer):
    viewer.current_focused_box = max(viewer.current_focused_box - 1, 0)
    viewer.verify_focused_box_in_view()


@bind(NORMAL, ALT_J, ALT_DOWN, description="Switch places with box below")
def switch_with_next_box(viewer):
    if viewer.current_focused_box + 1 >= viewer.num_boxes:
        return False
    viewer.swap_indices(viewer.current_focused_box, viewer.current_focused_box + 1)
    viewer.verify_focused_box_in_view()
    return FULL_REFRESH


@bind(NORMAL, ALT_K, ALT_UP, description="Switch places with box above")
def switch_with_previous_box(viewer):
    if viewer.current_focused_box - 1 < 0:
        return False
    viewer.swap_indices(viewer.current_focused_box, viewer.current_focused_box - 1)
    viewer.verify_focused_box_in_view()
    return FULL_REFRESH


@bind(NORMAL, "g", HOME, description="Focus first box")
def all_up(viewer):
    viewer.current_focused_box = 0
    viewer.verify_focused_box_in_view()


@bind(NORMAL, "G", END, description="Focus last box")
def all_down(viewer):
    viewer.current_focused_box = viewer.num_boxes - 1
    viewer.verify_focused_box_in_view()


@bind(NORMAL, "f", CTRL_F, PAGEDOWN, description="View 1 page down")
def page_down(viewer):
    viewer.current_view_line = min(viewer.max_current_line, viewer.current_view_line + viewer.lines)


@bind(NORMAL, "b", CTRL_B, PAGEUP, description="View 1 page up")
def page_up(viewer):
    viewer.current_view_line = max(0, viewer.current_view_line - viewer.lines)


@bind(NORMAL, "d", CTRL_D, description="View 1/2 page down")
def half_page_down(viewer):
    viewer.current_view_line = min(viewer.max_current_line, viewer.current_view_line + viewer.lines // 2)


@bind(NORMAL, "u", CTRL_U, description="View 1/2 page up")
def half_page_up(viewer):
    viewer.current_view_line = max(0, viewer.current_view_line - viewer.lines // 2)


@bind(NORMAL, CTRL_J, description="View 1 line down")
def line_down(viewer):
    viewer.current_view_line = min(viewer.max_current_line, viewer.current_view_line + 1)


@bind(NORMAL, CTRL_K, description="View 1 line up")
def line_up(viewer):
    viewer.current_view_line = max(0, viewer.current_view_line - 1)


# change box height
@bind(GLOBAL, ">", description="Increase box height")
def increase_box_height(viewer):
    return viewer.focused.increase_box_height()


@bind(GLOBAL, "<", description="Decrease box height")
def decrease_box_height(viewer):
    return viewer.focused.decrease_box_height()


# scroll mode
@bind(SCROLL, "j", DOWN, description="Scroll 1 line down")
def move_line_down(viewer):
    return viewer.focused.move_line_down()


@bind(SCROLL, "k", UP, description="Scroll 1 line up")
def move_line_up(viewer):
    return viewer.focused.move_line_up()


@bind(SCROLL, "g", HOME, description="Scroll to start")
def move_all_up(viewer):
    return viewer.focused.move_all_up()


@bind(SCROLL, "G", END, description="Scroll to end")
def move_all_down(viewer):
    return viewer.focused.move_all_down()


@bind(SCROLL, "b", CTRL_B, PAGEUP, description="Scroll 1 page up")
def move_page_up(viewer):
    return viewer.focused.move_page_up()


@bind(SCROLL, "f", CTRL_F, PAGEDOWN, description="Scroll 1 page down")
def move_page_down(viewer):
    return viewer.focused.move_page_down()


@bind(SCROLL, "u", CTRL_U, description="Scroll 1/2 page up")
def move_half_page_up(viewer):
    return viewer.focused.move_half_page_up()


@bind(SCROLL, "d", CTRL_D, description="Scroll 1/2 page down")
def move_half_page_down(viewer):
    return viewer.focused.move_half_page_down()


@bind(INPUT, CTRL__, description="Exit input mode")
def exit_input_mode(viewer):
    return viewer.focused.exit_input_mode()


# help mode
@bind(HELP, "j", DOWN, description="Scroll 1 line down [in help screen]")
def move_line_down(viewer):
    return viewer.help.move_line_down()


@bind(HELP, "k", UP, description="Scroll 1 line up [in help screen]")
def move_line_up(viewer):
    return viewer.help.move_line_up()



import asyncio
import logging
import types
import uuid
from dataclasses import dataclass
from typing import Any

import aiostream

from multiplex import ansi, to_iterator
from multiplex import keys
from multiplex import keys_input
from multiplex import resize
from multiplex import commands
from multiplex.actions import BoxAction
from multiplex.ansi import C, NONE
from multiplex.box import BoxHolder
from multiplex.enums import ViewLocation, BoxLine
from multiplex.exceptions import EndViewer
from multiplex.export import Export
from multiplex.help import HelpViewState
from multiplex.iterator import Descriptor
from multiplex.refs import REDRAW, RECALC, SPLIT, QUIT, ALL_DOWN, OUTPUT_SAVED, SAVE, STREAM_DONE

logger = logging.getLogger("multiplex.view")

MIN_BOX_HEIGHT = 7


class ViewerEvents:
    def __init__(self):
        self.queue = asyncio.Queue()

    async def receive(self):
        while True:
            yield await self.queue.get()
            self.queue.task_done()

    def send(self, message):
        self.queue.put_nowait(message)

    def send_redraw(self):
        self.queue.put_nowait((REDRAW, None))

    def send_recalc(self, num_boxes):
        self.queue.put_nowait((RECALC, num_boxes))

    def send_quit(self):
        self.queue.put_nowait((QUIT, None))

    def send_all_down(self):
        self.queue.put_nowait((ALL_DOWN, None))

    def send_save(self):
        self.queue.put_nowait((SAVE, None))

    def send_output_saved(self):
        self.queue.put_nowait((OUTPUT_SAVED, None))


@dataclass
class DescriptorQueueItem:
    descriptor: Descriptor
    redraw: bool
    num_boxes: int
    handle: Any = None


class Viewer:
    def __init__(self, descriptors, box_height, auto_collapse, verbose, socket_path, output_path, buffer_lines):
        self.holders = []
        self.stream_id_to_holder = {}
        self.holder_to_stream_id = {}
        self.loop = asyncio.get_event_loop()
        self.input_reader = keys_input.InputReader(viewer=self, bindings=keys.bindings)
        self.descriptors_queue: "asyncio.Queue[DescriptorQueueItem]" = asyncio.Queue()
        self.help = HelpViewState(self)
        self.events = ViewerEvents()
        self.export = Export(self)
        self.box_height = box_height
        self.auto_collapse = auto_collapse
        self.buffer_lines = buffer_lines
        self.verbose = verbose
        self.socket_path = socket_path
        self.output_path = output_path
        self.current_focused_box = 0
        self.current_view_line = 0
        self.maximized = False
        self.collaped_all = False
        self.wrapped_all = True
        self.cols = None
        self.lines = None
        self.stopped = False
        self.output_saved = False
        self.initial_add(descriptors)

    def initial_add(self, descriptors):
        for i, descriptor in enumerate(descriptors):
            redraw = i == 0 or i + 1 == len(descriptors)
            self.add(descriptor, redraw=redraw, num_boxes=len(descriptors))

    def add(self, descriptor, thread_safe=False, redraw=True, num_boxes=None):
        handle = None
        if descriptor.wait:
            handle = asyncio.Future()

        def action():
            self.descriptors_queue.put_nowait(
                DescriptorQueueItem(
                    descriptor=descriptor,
                    redraw=redraw,
                    num_boxes=num_boxes,
                    handle=handle,
                )
            )

        if thread_safe:
            self.loop.call_soon_threadsafe(action)
        else:
            action()

        return handle

    def split(self, title, box_height, stream_id):
        self.descriptors_queue.put_nowait(
            DescriptorQueueItem(
                descriptor=Descriptor(SPLIT, title, box_height, stream_id=stream_id),
                redraw=True,
                num_boxes=None,
            )
        )

    async def load(self, export_dir):
        await self.export.load(export_dir)

    def swap_indices(self, index1, index2):
        holder1 = self.get_holder(index1)
        holder2 = self.get_holder(index2)
        self.holders[index1] = holder2
        self.holders[index2] = holder1
        holder1.index = index2
        holder2.index = index1
        if self.current_focused_box == index1:
            self.current_focused_box = index2
        elif self.current_focused_box == index2:
            self.current_focused_box = index1

    @property
    def num_boxes(self):
        return len(self.holders)

    @property
    def iterators(self):
        return [h.iterator for h in self.holders]

    @property
    def buffers(self):
        return [h.buffer for h in self.holders]

    @property
    def states(self):
        return [h.state for h in self.holders]

    @property
    def boxes(self):
        return [h.box for h in self.holders]

    def get_holder(self, index):
        return self.holders[index]

    def get_buffer(self, index):
        return self.holders[index].buffer

    def get_state(self, index):
        return self.holders[index].state

    def get_iterator(self, index):
        return self.holders[index].iterator

    def get_box(self, index):
        return self.holders[index].box

    async def run(self):
        try:
            self._setup()
            await self._main()
        finally:
            self.restore()
            self.stopped = True

    def _setup(self):
        loop = asyncio.get_event_loop()
        keys_input.setup()
        resize_notifier = resize.setup(self.events, loop)
        ansi.setup()
        return resize_notifier

    @staticmethod
    def restore():
        loop = asyncio.get_event_loop()
        keys_input.restore()
        resize.restore(loop)
        ansi.restore()

    async def _main(self):
        self._init()
        async with aiostream.stream.advanced.flatten(self._sources()).stream() as streamer:
            async for obj, output in streamer:
                try:
                    await self._handle_event(obj, output)
                except EndViewer:
                    return

    async def _sources(self):
        yield self.input_reader.read()
        yield self.events.receive()
        while True:
            item = await self.descriptors_queue.get()
            source = await self._process_descriptor(item)
            if source:
                yield source
            self.descriptors_queue.task_done()

    async def _process_descriptor(self, descriptor_queue_item):
        descriptor = descriptor_queue_item.descriptor
        redraw = descriptor_queue_item.redraw
        recalc_num_boxes = descriptor_queue_item.num_boxes
        handle = descriptor_queue_item.handle
        index = self.num_boxes
        stream_id = str(uuid.uuid4())
        iterator = await to_iterator(
            obj=descriptor.obj,
            title=descriptor.title,
            context={
                "handle": handle,
                "socket_path": self.socket_path,
                "stream_id": stream_id,
                "cwd": descriptor.cwd,
                "env": descriptor.env,
            },
        )
        box_height = descriptor.box_height or self.box_height
        holder = BoxHolder(index, iterator=iterator, box_height=box_height, viewer=self)
        state = holder.state
        if descriptor.wrap is not None:
            state.wrap = descriptor.wrap
        if descriptor.collapsed is not None:
            state.collapsed = descriptor.collapsed
        self.holders.append(holder)
        if redraw:
            self.events.send_redraw()
        else:
            self.events.send_recalc(recalc_num_boxes)
        if redraw and not self.is_scrolling and descriptor.scroll_down:
            self.events.send_all_down()
        if iterator.iterator is SPLIT:
            if descriptor.stream_id:
                previous_holder = self.stream_id_to_holder[descriptor.stream_id]
            else:
                previous_holder = self.get_holder(index - 1)
            previous_holder.state.stream_done = True
            stream_id = self.holder_to_stream_id.pop(id(previous_holder))
        self.holder_to_stream_id[id(holder)] = stream_id
        self.stream_id_to_holder[stream_id] = holder
        if iterator.iterator is not SPLIT:
            return self._wrapped_iterator(stream_id, iterator.iterator)

    @staticmethod
    async def _wrapped_iterator(stream_id, iterator):
        async for elem in iterator:
            yield stream_id, elem
        yield stream_id, STREAM_DONE

    def _init(self):
        changed_cols, _ = self._update_lines_cols()
        if not self.holders:
            return
        ansi.clear()
        self._update_holders(changed_cols=changed_cols)
        self._update_view()
        ansi.flush()

    def _update_holders(self, num_boxes=None, changed_cols=None):
        num_boxes = num_boxes or self.num_boxes
        default_box_height = max(MIN_BOX_HEIGHT, (self.lines - num_boxes - 1) // num_boxes)
        for holder in self.holders:
            if changed_cols:
                holder.box.set_width(self.cols)
            if not holder.state.changed_height:
                holder.state.box_height = default_box_height

    def _update_lines_cols(self):
        cols, lines = ansi.get_size()
        prev_cols = self.cols
        prev_lines = self.lines
        self.cols = cols
        self.lines = lines
        changed_cols = prev_cols != self.cols
        changed_lines = prev_lines != self.lines
        if changed_cols or changed_lines:
            logger.debug(f"sizes: prev [{prev_lines}, {prev_cols}], new [{self.lines}, {self.cols}]")
        return changed_cols, changed_lines

    async def _handle_event(self, obj, output):
        if obj is REDRAW:
            self._init()
            return
        if obj is RECALC:
            self._update_holders(output)
            return
        if obj is QUIT:
            raise EndViewer

        if obj is SAVE:
            await commands.save(self)
            self._update_status_bar()
        elif obj is OUTPUT_SAVED:
            await asyncio.sleep(0.1)
            self.output_saved = False
            self._update_status_bar()
        elif obj is ALL_DOWN:
            commands.all_down(self)
            ansi.clear()
            self._update_view()
        elif isinstance(obj, str):
            holder = self.stream_id_to_holder[obj]
            self._update_box(holder.index, output)
            if isinstance(output, BoxAction) or callable(output) or output is STREAM_DONE:
                ansi.clear()
                self._update_view()
            if self.is_input_mode:
                self._update_cursor()
        else:
            key_changed = False
            boxes_changed = set()
            full_refresh = False
            for fn in output:
                current_key_changed = await self._process_key_handler(fn)
                if current_key_changed is ansi.FULL_REFRESH:
                    full_refresh = True
                    key_changed = True
                elif type(current_key_changed) == int:
                    boxes_changed.add(current_key_changed)
                else:
                    key_changed = key_changed or current_key_changed
            if full_refresh:
                ansi.clear()
            if key_changed:
                self._update_view()
            elif boxes_changed:
                self._update_status_bar()
                for index in boxes_changed:
                    self._update_box(index, data=None)
            elif not self.is_input_mode:
                self._update_status_bar()
        ansi.flush()

    def _update_box(self, i, data):
        if data is not None:
            holder = self.get_holder(i)
            if isinstance(data, BoxAction):
                data.run(holder)
            elif callable(data):
                data(holder)
            elif data is STREAM_DONE:
                holder.state.stream_done = True
                if self.auto_collapse and not holder.iterator.metadata.get("exit_code"):
                    holder.box.toggle_collapse(value=True)
                holder.box.exit_input_mode()
            else:
                holder.buffer.write(data)
        if self.help.show:
            return
        self._update_title_line(i)
        box = self.get_box(i)
        if box.is_visible:
            box.update()

    async def _process_key_handler(self, fn):
        prev_current_line = self.current_view_line
        prev_focused_box = self.current_focused_box
        update_view = fn(self)
        if isinstance(update_view, types.CoroutineType):
            update_view = await update_view
        if update_view is not None and not isinstance(update_view, bool):
            return update_view
        if prev_focused_box != self.current_focused_box:
            return ansi.FULL_REFRESH
        if prev_current_line != self.current_view_line:
            return ansi.FULL_REFRESH
        return update_view

    def _update_view(self):
        if self.help.show:
            ansi.help_screen(
                current_line=self.help.current_line,
                lines=self.lines,
                cols=self.cols - 1,
                descriptions=keys.descriptions,
            )
        else:
            self._update_status_bar()
            self._update_title_lines()
            self._update_boxes()
            self._update_cursor()

    def _update_boxes(self):
        if self.maximized:
            self._update_box(self.current_focused_box, data=None)
        else:
            for i in range(self.num_boxes):
                self._update_box(i, data=None)

    def _update_title_lines(self):
        if self.maximized:
            self._update_title_line(self.current_focused_box)
        else:
            for i in range(self.num_boxes):
                self._update_title_line(i)

    def _update_title_line(self, index):
        screen_y, location = self.get_title_line(index)
        if location != ViewLocation.IN_VIEW:
            return

        holder = self.get_holder(index)
        iterator = holder.iterator
        box_state = holder.state
        suffix = ""
        if self.verbose:
            wrap = box_state.wrap
            auto_scroll = box_state.auto_scroll
            collapsed = box_state.collapsed
            input_mode = box_state.input_mode
            buffer_line = box_state.buffer_start_line
            box_height = box_state.box_height
            state = [
                "W" if wrap else "-",
                "-" if auto_scroll else "S",
                "C" if collapsed else "-",
                "I" if input_mode else "-",
            ]
            state = f"{''.join(state)} [{buffer_line},{box_height}]"
            suffix = f" [{state}]"
        suffix_len = len(suffix)
        title = iterator.title
        if not isinstance(title, C):
            title = C(title)
        title_len = len(title)
        hr_space = 4
        _ellipsis = "..."
        if hr_space + title_len + suffix_len > self.cols:
            title = title[: self.cols - suffix_len - len(_ellipsis) - hr_space]
            title += _ellipsis
        text = C(" ", title, suffix, " ", color=(NONE, NONE))

        logger.debug(f"s{index}:\t{screen_y}\t{location}\t[{self.lines},{self.cols}]")

        if index == self.current_focused_box:
            hline_color = ansi.theme.TITLE_FOCUS
        elif box_state.stream_done:
            hline_color = ansi.theme.TITLE_STREAM_DONE
        else:
            hline_color = ansi.theme.TITLE_NORMAL
        ansi.title(
            row=screen_y,
            text=text,
            hline_color=hline_color,
            cols=self.cols,
        )

    def _update_status_bar(self):
        if self.help.show:
            return

        focused = self.focused
        index = focused.index
        iterator = self.get_iterator(index)
        title = iterator.title
        box_state = focused.state
        wrap = box_state.wrap
        auto_scroll = box_state.auto_scroll
        input_mode = box_state.input_mode

        modes = []
        if not auto_scroll:
            modes.append("SCROLL")
        if self.maximized:
            modes.append("MAXIMIZED")
        if wrap:
            modes.append("WRAP")
        if input_mode:
            modes.append("INPUT")
        mode = "|".join(modes)
        mode_paren = f"({mode})" if mode else ""

        pending_keys = self.input_reader.pending
        pending_name_parts = []
        while pending_keys:
            cur_offset = len(pending_keys)
            name = None
            while not name and cur_offset:
                name = keys.seq_to_name(pending_keys[:cur_offset], fallback=False)
                if not name:
                    cur_offset -= 1
            pending_name_parts.append(name)
            pending_keys = pending_keys[cur_offset:]

        pending_text = "".join(pending_name_parts)
        if pending_text:
            pending_text = f"{pending_text} "

        prefix = f"[{index + 1}] "
        prefix_len = len(prefix)

        if not isinstance(title, C):
            title = C(title)
        title_len = len(title)
        mode_len = len(mode_paren)
        pending_len = len(pending_text)
        space_between = self.cols - title_len - mode_len - pending_len - prefix_len
        if space_between < 0:
            _ellipsis = "... "
            title = title[: (self.cols - mode_len - pending_len - prefix_len) - len(_ellipsis)]
            title += _ellipsis
            space_between = 0
        if self.output_saved:
            color = ansi.theme.STATUS_SAVE
        elif input_mode:
            color = ansi.theme.STATUS_INPUT
        elif not auto_scroll:
            color = ansi.theme.STATUS_SCROLL
        else:
            color = ansi.theme.STATUS_NORMAL
        text = C(prefix, title, " " * space_between, pending_text, mode_paren, color=color)

        ansi.status_bar(
            row=self.get_status_bar_line(),
            text=text,
        )

    def _update_cursor(self):
        if not self.is_input_mode:
            return
        box = self.focused
        screen_y, view_location = self.get_box_top_line(box.index)
        if view_location != ViewLocation.IN_VIEW:
            return
        col, row = box.buffer.get_cursor(box.state.wrap)
        row += screen_y
        row -= box.state.buffer_start_line
        if row < 0 or col < 0:
            return
        ansi.move_cursor(col, row)

    def verify_focused_box_in_view(self):
        if self.maximized:
            return
        screen_y, location = self.get_box_bottom_line(self.current_focused_box)
        if location == ViewLocation.BELOW:
            offset = screen_y
            self.current_view_line += offset
            return
        screen_y, location = self.get_title_line(self.current_focused_box)
        if location == ViewLocation.ABOVE:
            offset = screen_y
            self.current_view_line -= offset
        elif location == ViewLocation.BELOW:
            offset = screen_y
            self.current_view_line += offset

    @property
    def focused(self):
        return self.get_box(self.current_focused_box)

    @property
    def is_scrolling(self):
        return not self.focused.state.auto_scroll

    @property
    def is_input_mode(self):
        return self.focused.state.input_mode

    @property
    def max_current_line(self):
        result = 0
        for state in self.states:
            result += 1
            if not state.collapsed:
                result += state.box_height
        result -= self.get_max_box_line()
        return max(0, result)

    def get_status_bar_line(self):
        return self.lines - 1

    def get_max_box_line(self):
        return self.lines - 2

    def get_title_line(self, index):
        return self._get_box_line(index, BoxLine.TITLE)

    def get_box_top_line(self, index):
        return self._get_box_line(index, BoxLine.TOP)

    def get_box_bottom_line(self, index):
        return self._get_box_line(index, BoxLine.BOTTOM)

    def _get_box_line(self, index, box_line):
        if self.maximized:
            if index != self.current_focused_box:
                return 0, ViewLocation.NOT_FOCUSED
            if box_line == BoxLine.TITLE:
                return 0, ViewLocation.NOT_FOCUSED
            elif box_line == BoxLine.TOP:
                screen_y = 0
            elif box_line == BoxLine.BOTTOM:
                screen_y = self.get_max_box_line()
            else:
                raise RuntimeError(box_line)
            return screen_y, ViewLocation.IN_VIEW
        else:
            state = self.get_state(index)
            if state.collapsed and box_line != BoxLine.TITLE:
                return 0, ViewLocation.NOT_FOCUSED
            view_y = 0
            for i in range(index):
                # title line
                view_y += 1
                other_state = self.states[i]
                if not other_state.collapsed:
                    view_y += other_state.box_height
            if box_line == BoxLine.TOP:
                view_y += 1
            elif box_line == BoxLine.BOTTOM:
                view_y += state.box_height
            screen_y = view_y - self.current_view_line
            max_line = self.get_max_box_line()
            if screen_y > max_line:
                offest = screen_y - max_line
                return offest, ViewLocation.BELOW
            elif screen_y < 0:
                offset = abs(screen_y)
                return offset, ViewLocation.ABOVE
            else:
                return screen_y, ViewLocation.IN_VIEW



import asyncio
import fcntl
import io
import json
import os
import struct
import termios
import time
import types
import pathlib
import pty
from dataclasses import dataclass
from typing import Any

import aiofiles
from aiostream.stream import create, combine
from multiplex import ansi
from multiplex.actions import SetTitle, BoxActions, UpdateMetadata
from multiplex.ansi import C, theme
from multiplex.controller import Controller
from multiplex.process import Process
from multiplex.refs import SPLIT, STOP

MULTIPLEX_SOCKET_PATH = "MULTIPLEX_SOCKET_PATH"
MULTIPLEX_STREAM_ID = "MULTIPLEX_STREAM_ID"


async def stream_reader_generator(reader):
    while True:
        try:
            b = await reader.read(1000000)
            if not b:
                break
            yield b.decode()
        except OSError:
            return


async def asciinema_recording_iterator(recording_path):
    async with aiofiles.open(recording_path, encoding="utf-8") as f:
        after_first_line = False
        last_abs_time = time.time()
        last_rel_time = 0
        async for line in f:
            if not after_first_line:
                after_first_line = True
                continue
            rel_time, type_, output = json.loads(line)
            if type_ != "o":
                continue
            abs_time = time.time()
            sleep_time = (rel_time - last_rel_time) - (abs_time - last_abs_time)
            await asyncio.sleep(sleep_time)
            last_rel_time = rel_time
            last_abs_time = abs_time
            yield output


@dataclass
class Descriptor:
    obj: Any
    title: str
    box_height: int
    wrap: bool = None
    collapsed: bool = None
    scroll_down: bool = False
    # only relevant for ipc requests
    wait: bool = False
    stream_id: str = None
    cwd: str = None
    env: dict = None


class Iterator:
    def __init__(self, iterator, title, inner_type, metadata):
        self.iterator = iterator
        self.title = title
        self.inner_type = inner_type
        self.metadata = metadata


def _extract_title(current_title, obj):
    if current_title is not None:
        return current_title
    if isinstance(obj, str):
        if obj.startswith("file://") or obj.startswith("asciinema://"):
            return obj.split("://")[1]
        return obj
    if isinstance(obj, pathlib.Path):
        return str(obj)
    title_attr = getattr(obj, "title", None)
    if title_attr:
        return title_attr
    name_attr = getattr(obj, "__name__", None)
    if name_attr:
        return name_attr
    class_attr = getattr(obj, "__class__", None)
    if class_attr:
        name_attr = getattr(class_attr, "__name__", None)
        if name_attr:
            return name_attr
    return None


def _process_descriptor_to_iterator(descriptor: Process, context):
    def _setsize(fd):
        cols, rows = ansi.get_size()
        s = struct.pack("HHHH", rows, cols, 0, 0)
        fcntl.ioctl(fd, termios.TIOCSWINSZ, s)

    master, slave = pty.openpty()
    _setsize(slave)
    cwd = context.get("cwd")
    env = (context.get("env") or os.environ).copy()
    env.update(
        {
            MULTIPLEX_SOCKET_PATH: context.get("socket_path", ""),
            MULTIPLEX_STREAM_ID: context.get("stream_id", ""),
        }
    )
    cmd = descriptor.cmd
    if isinstance(cmd, str):
        obj = asyncio.subprocess.create_subprocess_shell(
            cmd,
            stdin=slave,
            stdout=slave,
            stderr=slave,
            env=env,
            cwd=cwd,
        )
    else:
        obj = asyncio.subprocess.create_subprocess_exec(
            cmd[0],
            *cmd[1:],
            stdin=slave,
            stdout=slave,
            stderr=slave,
            env=env,
            cwd=cwd,
        )
    return obj, (master, slave)


def _str_to_iterator(str_value, title, context):
    title = _extract_title(title, str_value)
    master, slave = None, None
    if str_value.startswith("asciinema://"):
        obj = asciinema_recording_iterator(str_value.split("://", 1)[1])
    elif str_value.startswith("file://"):
        obj = pathlib.Path(str_value.split("://")[1])
    else:
        obj, (master, slave) = _process_descriptor_to_iterator(Process(str_value), context)
    return obj, title, (master, slave)


def _coroutine_to_iterator(cor):
    loop = asyncio.get_event_loop()
    if loop.is_running():
        raise RuntimeError("Event loop is current running. Use the async version")
    return loop.run_until_complete(cor)


def _process_to_iterator(process, title, master, slave, context):
    handle: asyncio.Future = context.get("handle")

    async def exit_stream():
        exit_code = await process.wait()
        if handle:
            if exit_code:
                handle.set_exception(RuntimeError(f"'{title}' failed with code {exit_code}"))
            else:
                handle.set_result(True)
        if slave:
            os.close(slave)
        yield

    async def g():
        stdout = process.stdout
        stderr = process.stderr
        streams = []
        if isinstance(stdout, asyncio.streams.StreamReader):
            streams.append(stream_reader_generator(stdout))
        if isinstance(stderr, asyncio.streams.StreamReader):
            streams.append(stream_reader_generator(stderr))
        if master:
            loop = asyncio.get_running_loop()
            assert slave
            pipe = io.open(master, "w+b", 0)
            reader = asyncio.StreamReader()
            reader_protocol = asyncio.StreamReaderProtocol(reader)
            await loop.connect_read_pipe(lambda: reader_protocol, pipe)
            streams.extend([stream_reader_generator(reader), exit_stream()])
            writer_transport, writer_protocol = await loop.connect_write_pipe(asyncio.streams.FlowControlMixin, pipe)
            writer = asyncio.StreamWriter(writer_transport, writer_protocol, reader, loop)
            yield UpdateMetadata({"input": writer})

        assert streams
        stream = combine.merge(*streams)
        async with stream.stream() as s:
            async for data in s:
                yield data
        exit_code = process.returncode if slave else await process.wait()
        status = C("✗", color=theme.X_COLOR) if exit_code else C("✓", color=theme.V_COLOR)
        yield BoxActions(
            [
                UpdateMetadata({"exit_code": exit_code}),
                SetTitle(C("[", status, f"] {title}")),
            ]
        )

    obj = g()
    title = _extract_title(title, process)
    inner_type = "async_process"

    def close():
        try:
            process.kill()
        except ProcessLookupError:
            pass

    return obj, title, inner_type, {"close": close}


def _path_to_iterator(file_path, title):
    async def g():
        async with aiofiles.open(file_path, encoding="utf-8") as f:
            yield await f.read()

    obj = g()
    title = _extract_title(title, file_path)
    inner_type = "path"
    return obj, title, inner_type


def _controller_to_iterator(controller, title):
    async def g():
        controller._init()
        while True:
            result = await controller.queue.get()
            if result is STOP:
                controller.queue.task_done()
                break
            yield result
            controller.queue.task_done()

    obj = g()
    title = _extract_title(title, controller)
    inner_type = "controller"
    return obj, title, inner_type


def _stream_reader_to_iterator(stream_reader, title):
    obj = stream_reader_generator(stream_reader)
    title = _extract_title(title, stream_reader)
    inner_type = "stream_reader"
    return obj, title, inner_type


def _async_generator_to_iterator(agen, title):
    obj = agen
    title = _extract_title(title, agen)
    inner_type = "async_generator"
    return obj, title, inner_type


def _generator_to_iterator(gen, title):
    obj = create.from_iterable.raw(gen)
    title = _extract_title(title, gen)
    inner_type = "generator"
    return obj, title, inner_type


def _async_iterable_to_iterator(aiter, title):
    obj = aiter
    title = _extract_title(title, aiter)
    inner_type = "async_iterable"
    return obj, title, inner_type


def _iterable_to_iterator(_iter, title):
    obj = create.from_iterable.raw(_iter)
    title = _extract_title(title, _iter)
    inner_type = "iterable"
    return obj, title, inner_type


def _callable_to_iterator(cb, title):
    obj = cb()
    title = _extract_title(title, cb)
    return obj, title


async def _to_iterator(obj, title, context):
    master, slave = None, None
    if isinstance(obj, str):
        obj, title, (master, slave) = _str_to_iterator(obj, title, context)
    elif isinstance(obj, Process):
        title = _extract_title(title, obj)
        obj, (master, slave) = _process_descriptor_to_iterator(obj, context)
    elif isinstance(obj, (types.FunctionType, types.MethodType)):
        obj, title = _callable_to_iterator(obj, title)

    if isinstance(obj, types.CoroutineType):
        obj = await obj

    if isinstance(obj, asyncio.streams.StreamReader):
        result = _stream_reader_to_iterator(obj, title)
    elif isinstance(obj, types.AsyncGeneratorType):
        result = _async_generator_to_iterator(obj, title)
    elif isinstance(obj, types.GeneratorType):
        result = _generator_to_iterator(obj, title)
    elif hasattr(obj, "__aiter__"):
        result = _async_iterable_to_iterator(obj, title)
    elif hasattr(obj, "__iter__"):
        result = _iterable_to_iterator(obj, title)
    elif isinstance(obj, asyncio.subprocess.Process):
        result = _process_to_iterator(obj, title, master, slave, context)
    elif isinstance(obj, pathlib.Path):
        result = _path_to_iterator(obj, title)
    elif isinstance(obj, Controller):
        result = _controller_to_iterator(obj, title)
    else:
        raise RuntimeError(f"Cannot create iterator from {obj}")

    metadata = {}
    if len(result) == 3:
        obj, title, inner_type = result
    else:
        obj, title, inner_type, metadata = result

    if title is None:
        title = "N/A"

    return obj, title, inner_type, metadata


async def to_iterator(obj, title=None, context=None) -> Iterator:
    if obj is SPLIT:
        return Iterator(SPLIT, title, "split", {})
    return Iterator(*(await _to_iterator(obj, title, context or {})))



import signal


def setup(viewer_events, loop):
    def sigwinch():
        viewer_events.send_redraw()

    loop.add_signal_handler(signal.SIGWINCH, sigwinch)


def restore(loop):
    loop.remove_signal_handler(signal.SIGWINCH)



import asyncio
import json
import os
import tempfile
from asyncio import StreamWriter, StreamReader
from random import randint

from multiplex.exceptions import IPCException
from multiplex.viewer import Viewer
from multiplex.iterator import Descriptor, MULTIPLEX_STREAM_ID


class Server:
    def __init__(self, socket_path=None):
        self.socket_path = socket_path or f"{tempfile.gettempdir()}/multiplex-{randint(100000, 999999)}"
        self.viewer: Viewer = None
        self.server = None
        self.server_task = None
        self.stopped = False

    async def start(self, viewer):
        self.viewer = viewer
        self.server = await asyncio.start_unix_server(
            client_connected_cb=self._handle_request,
            path=self.socket_path,
        )
        self.server_task = asyncio.create_task(self.server.serve_forever())

    def stop(self):
        if self.server_task:
            self.server_task.cancel()
        if os.path.exists(self.socket_path):
            os.remove(self.socket_path)
        self.stopped = True

    async def _handle_request(self, reader: StreamReader, writer: StreamWriter):
        request = await _read_message(reader)
        try:
            await self._handle_message(request)
            response = {"status": "success"}
        except Exception as e:
            response = {"status": "failure", "error": str(e)}
        await _write_message(response, writer)
        writer.close()
        return response

    async def _handle_message(self, message, batch=False):
        action = message.pop("action")
        if action == "split":
            self.viewer.split(**message)
        elif action == "collapse":
            self.viewer.focused.toggle_collapse(**message)
            self.viewer.events.send_redraw()
        elif action == "add":
            descriptor = Descriptor(**message, scroll_down=True)
            handle = self.viewer.add(descriptor)
            if descriptor.wait:
                if batch:
                    return handle
                else:
                    await handle
        elif action == "save":
            self.viewer.events.send_save()
        elif action == "load":
            await self.viewer.load(**message)
        elif action == "quit":
            self.viewer.events.send_quit()
        elif action == "batch":
            handles = []
            for action in message["actions"]:
                handle = self._handle_message(action)
                if handle:
                    handles.append(handle)
            if handles:
                await asyncio.gather(*handles)


class Client:
    def __init__(self, socket_path):
        self._socket_path = socket_path

    async def add(self, obj, title, box_height, wait, cwd, env):
        await self._request(self.add_request_body(obj, title, box_height, wait, cwd, env))

    @staticmethod
    def add_request_body(obj, title, box_height, wait, cwd, env):
        return {
            "action": "add",
            "obj": obj,
            "title": title,
            "box_height": box_height,
            "wait": wait,
            "cwd": cwd,
            "env": env,
        }

    async def split(self, title, box_height, stream_id):
        await self._request(self.split_request_body(title, box_height, stream_id))

    @staticmethod
    def split_request_body(title, box_height, stream_id):
        return {"action": "split", "title": title, "box_height": box_height, "stream_id": stream_id}

    async def toggle_collapse(self, value=None):
        await self._request(self.collapse_request_body(value))

    @staticmethod
    def collapse_request_body(value):
        return {"action": "collapse", "value": value}

    async def save(self):
        await self._request(self.save_request_body())

    @staticmethod
    def save_request_body():
        return {"action": "save"}

    async def load(self, export_dir):
        await self._request(self.load_request_body(export_dir))

    @staticmethod
    def load_request_body(export_dir):
        return {"action": "load", "export_dir": export_dir}

    async def quit(self):
        await self._request(self.quit_request_body())

    @staticmethod
    def quit_request_body():
        return {"action": "quit"}

    async def batch(self, actions):
        await self._request({"action": "batch", "actions": actions})

    async def _connect(self):
        return await asyncio.open_unix_connection(self._socket_path)

    async def _request(self, message):
        reader, writer = await self._connect()
        await _write_message(message, writer)
        response = await _read_message(reader)
        writer.close()
        if response["status"] == "failure":
            raise IPCException(response["error"])
        return response


async def _read_message(reader):
    return json.loads((await reader.readline()).decode().strip())


async def _write_message(message, writer):
    writer.write(f"{json.dumps(message)}\n".encode())
    await writer.drain()


def get_env_stream_id():
    return os.environ.get(MULTIPLEX_STREAM_ID)



import asyncio
import os
from typing import List

from multiplex.viewer import Viewer
from multiplex.iterator import Descriptor
from multiplex.ipc import Server


class Multiplex:
    def __init__(
        self, verbose=False, box_height=None, auto_collapse=False, output_path=None, socket_path=None, buffer_lines=None
    ):
        self.descriptors: List[Descriptor] = []
        self.verbose = verbose
        self.box_height = box_height
        self.auto_collapse = auto_collapse
        self.buffer_lines = buffer_lines
        self.output_path = output_path or os.getcwd()
        self.server = Server(socket_path)
        self.viewer: Viewer = None

    def run(self, load=None):
        try:
            asyncio.run(self.run_async(load))
        except KeyboardInterrupt:
            pass
        finally:
            self.cleanup()

    async def run_async(self, load=None):
        assert not self.viewer
        self.viewer = Viewer(
            descriptors=self.descriptors,
            verbose=self.verbose,
            box_height=self.box_height,
            auto_collapse=self.auto_collapse,
            socket_path=self.server.socket_path,
            output_path=self.output_path,
            buffer_lines=self.buffer_lines,
        )
        if load:
            await self.viewer.load(load)
        await self.server.start(viewer=self.viewer)
        try:
            await self.viewer.run()
        finally:
            self.server.stop()

    def add(self, obj, title=None, box_height=None, thread_safe=False):
        descriptor = Descriptor(obj=obj, title=title, box_height=box_height, scroll_down=self.viewer is not None)
        self.descriptors.append(descriptor)
        if self.viewer:
            self.viewer.add(descriptor, thread_safe=thread_safe)

    def add_thread_safe(self, obj, title=None, box_height=None):
        self.add(obj, title, box_height, thread_safe=True)

    def cleanup(self):
        if self.viewer:
            if not self.viewer.stopped:
                self.viewer.restore()
            for it in self.viewer.iterators:
                close = it.metadata.get("close")
                if close:
                    close()
        if self.server and not self.server.stopped:
            self.server.stop()

    @property
    def socket_path(self):
        return self.server.socket_path



import logging
import os
import tempfile


def init_logging(level=logging.INFO, log_path=None):
    formatter = logging.Formatter("{asctime} - {name:<20} - {levelname} - {message}", style="{")
    log_path = log_path or os.path.join(tempfile.gettempdir(), "multiplex.log")
    fh = logging.FileHandler(log_path)
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger = logging.getLogger()
    for h in logger.handlers:
        logger.removeHandler(h)
    logger.setLevel(level)
    logger.addHandler(fh)



from contextlib import contextmanager
from functools import partial
from unittest import mock
from unittest.mock import Mock

from multiplex.keys import bind as _bind, HELP, SCROLL, NORMAL, GLOBAL, INPUT
from multiplex.keys_input import InputReader


def _test_set(show_help=False, auto_scroll=True, input_mode=False):
    bindings = {GLOBAL: {}, NORMAL: {}, HELP: {}, SCROLL: {}, INPUT: {}}
    descriptions = {GLOBAL: {}, NORMAL: {}, HELP: {}, SCROLL: {}, INPUT: {}}
    bind = partial(_bind, custom_bindings=bindings, custom_descriptions=descriptions)
    viewer = MockViewer(show_help, auto_scroll, input_mode)
    return bind, InputReader(viewer, bindings)


class MockViewer:
    def __init__(self, show_help=False, auto_scroll=True, input_mode=False):
        self.help = MockHelp(show_help)
        self.is_input_mode = input_mode
        self.is_scrolling = not auto_scroll


class MockHelp:
    def __init__(self, show=False):
        self.show = show


class MockBox:
    def __init__(self, auto_scroll):
        self.state = MockBoxState(auto_scroll)
        pass


class MockBoxState:
    def __init__(self, auto_scroll):
        self.auto_scroll = auto_scroll


def _test_process_single_found_sequence(mode):
    show_help = mode == HELP
    auto_scroll = mode != SCROLL
    input_mode = mode == INPUT
    bind, reader = _test_set(show_help, auto_scroll, input_mode)

    @bind(mode, "a")
    def fn1():
        pass

    expected = [fn1]

    if mode != GLOBAL:

        @bind(GLOBAL, "a")
        def fn2():
            pass

        @bind(GLOBAL, "b")
        def fn3():
            pass

        if mode != INPUT:
            expected.append(fn3)

    assert reader._process([ord("a"), ord("b")]) == (expected, [])


def test_process_single_found_sequence_normal():
    _test_process_single_found_sequence(NORMAL)


def test_process_single_found_sequence_scroll():
    _test_process_single_found_sequence(SCROLL)


def test_process_single_found_sequence_input():
    _test_process_single_found_sequence(INPUT)


def test_process_single_found_sequence_help():
    _test_process_single_found_sequence(HELP)


def test_process_single_found_sequence_global():
    _test_process_single_found_sequence(GLOBAL)


def test_process_pending():
    bind, reader = _test_set()

    @bind(NORMAL, "ab")
    def fn1():
        pass

    keys1 = [ord("a")]
    keys2 = [ord("a"), ord("b")]
    keys3 = [ord("a"), ord("c"), ord("d")]
    assert reader._process(keys1) == ([], keys1)
    assert reader._process(keys2) == ([fn1], [])
    assert reader._process(keys3) == ([], [])

    @bind(NORMAL, "def")
    def fn2():
        pass

    keys1 = [ord("d")]
    keys2 = [ord("d"), ord("b")]
    keys3 = [ord("d"), ord("e")]
    assert reader._process(keys1) == ([], keys1)
    assert reader._process(keys2) == ([], [])
    assert reader._process(keys3) == ([], keys3)


def test_process_more_than_one_sequence():
    bind, reader = _test_set()

    @bind(NORMAL, "ab")
    def fn1():
        pass

    @bind(NORMAL, "cd")
    def fn2():
        pass

    keys1 = [ord("a"), ord("b"), ord("a")]
    keys2 = [ord("a"), ord("b"), ord("a"), ord("b")]
    keys3 = [ord("a"), ord("b"), ord("c"), ord("d")]
    assert reader._process(keys1) == ([fn1], keys1[2:])
    assert reader._process(keys2) == ([fn1, fn1], [])
    assert reader._process(keys3) == ([fn1, fn2], [])


@contextmanager
def patch_read(data=None, read=None):
    mock_has_data = Mock()
    mock_has_data.side_effect = data or [False]
    mock_read = Mock()
    mock_read.side_effect = read or []
    with mock.patch("sys.stdin.fileno", Mock()):
        with mock.patch("os.read", mock_read):
            with mock.patch("multiplex.keys_input._has_data", mock_has_data):
                yield


def test_read_iteration_no_data():
    bind, reader = _test_set()
    assert reader.pending == []
    with patch_read(data=[False]):
        assert reader._read_iteration() is None
    assert reader.pending == []


def test_read_iteration_process_one():
    bind, reader = _test_set()

    @bind(NORMAL, "a")
    def fn1():
        pass

    with patch_read(data=[True, True, False], read=["a"]):
        assert reader._read_iteration() == [fn1]

    assert reader.pending == []


def test_read_iteration_has_pending():
    bind, reader = _test_set()

    @bind(NORMAL, "ab")
    def fn1():
        pass

    with patch_read(data=[True, True, False], read=["a"]):
        assert reader._read_iteration() == []

    assert reader.pending == [ord("a")]


def test_read_iteration_backspace_and_delete():
    bind, reader = _test_set()

    @bind(NORMAL, "abcdef")
    def fn1():
        pass

    with patch_read(data=[True, True, True, True, True, True, True, True], read=["a", "b", "c", chr(8), chr(127)]):
        assert reader._read_iteration() == []
        assert reader.pending == [ord("a"), ord("b")]
        assert reader._read_iteration() == []
        assert reader.pending == [ord("a")]



import json
import os
import string
from datetime import datetime

import aiofiles
from multiplex.ansi import C
from multiplex.iterator import Descriptor


class Export:
    def __init__(self, viewer):
        self.viewer = viewer

    async def save(self):
        viewer = self.viewer
        now = datetime.now()
        dir_name = f"output-{now.strftime('%Y-%m-%dT%H-%M-%S')}"
        output_dir = os.path.join(viewer.output_path, dir_name)
        os.makedirs(output_dir, exist_ok=True)
        valid_chars = string.ascii_letters + string.digits
        holders = viewer.holders
        zero_padding = 1 if len(holders) < 10 else 2
        metadata = {
            "view": {
                "maximized": viewer.maximized,
                "collapsed": viewer.collaped_all,
                "wrapped": viewer.wrapped_all,
            },
            "boxes": [],
        }
        for index, holder in enumerate(holders):
            initial_title = holder.iterator.title
            state = holder.state
            title = initial_title.to_string(no_style=True) if isinstance(initial_title, C) else str(initial_title)
            title = "".join(c for c in title if c in valid_chars).lower()
            file_name = f"{str(index + 1).zfill(zero_padding)}-{title}"
            async with aiofiles.open(os.path.join(output_dir, file_name), "w") as f:
                await f.write(holder.buffer.raw_buffer.getvalue())
            metadata["boxes"].append(
                {
                    "title": initial_title.to_dict() if isinstance(initial_title, C) else initial_title,
                    "box_height": state.box_height,
                    "collapsed": state.collapsed,
                    "wrap": state.wrap,
                    "filename": file_name,
                }
            )
        async with aiofiles.open(os.path.join(output_dir, "metadata.json"), "w") as f:
            await f.write(json.dumps(metadata, indent=2))
        viewer.output_saved = True
        viewer.events.send_output_saved()

    async def load(self, export_dir):
        viewer = self.viewer
        async with aiofiles.open(os.path.join(export_dir, "metadata.json")) as f:
            metadata = json.loads(await f.read())
        viewer.initial_add(
            [
                Descriptor(
                    obj=f"file://{export_dir}/{box['filename']}",
                    title=C.from_dict(box["title"]) if isinstance(box["title"], dict) else box["title"],
                    box_height=box["box_height"],
                    collapsed=box["collapsed"],
                    wrap=box["wrap"],
                )
                for box in metadata["boxes"]
            ]
        )
        view = metadata["view"]
        viewer.maximized = view["maximized"]
        viewer.collaped_all = view["collapsed"]
        viewer.wrapped_all = view["wrapped"]



BACKSPACE = 8
DEL = 127
BACKSPACE_OR_DEL = {BACKSPACE, DEL}

ESC = 27
ESC_MOVE = [ESC, 91]
CTRL_MOVE = ESC_MOVE + [49, 59, 53]

UP = ESC_MOVE + [65]
DOWN = ESC_MOVE + [66]
LEFT = ESC_MOVE + [67]
RIGHT = ESC_MOVE + [68]

HOME = ESC_MOVE + [72]
END = ESC_MOVE + [70]

PAGEUP = ESC_MOVE + [53, 126]
PAGEDOWN = ESC_MOVE + [54, 126]

CTRL_UP = CTRL_MOVE + [65]
CTRL_DOWN = CTRL_MOVE + [66]
CTRL_LEFT = CTRL_MOVE + [67]
CTRL_RIGHT = CTRL_MOVE + [68]

ALT_UP = [ESC] + UP
ALT_DOWN = [ESC] + DOWN
ALT_LEFT = [ESC] + LEFT
ALT_RIGHT = [ESC] + RIGHT

ALT_J = [ESC] + [ord("j")]
ALT_K = [ESC] + [ord("k")]

CTRL_K = (11,)
CTRL_J = (10,)
CTRL_B = (2,)
CTRL_F = (6,)
CTRL_U = (21,)
CTRL_D = (4,)
CTRL__ = (27,)

CTRL_TO_NAME = {
    CTRL_K: "K",
    CTRL_J: "J",
    CTRL_B: "B",
    CTRL_F: "F",
    CTRL_U: "U",
    CTRL_D: "D",
}

UP_NAME = "↑"
DOWN_NAME = "↓"
LEFT_NAME = "←"
RIGHT_NAME = "→"
ALT_NAME = "⌥"

SEQ_TO_NAME = {
    tuple(UP): UP_NAME,
    tuple(DOWN): DOWN_NAME,
    tuple(LEFT): LEFT_NAME,
    tuple(RIGHT): RIGHT_NAME,
    tuple(HOME): "Home",
    tuple(END): "End",
    tuple(PAGEUP): "Page Up",
    tuple(PAGEDOWN): "Page Down",
    tuple(CTRL_UP): f"^{UP_NAME}",
    tuple(CTRL_DOWN): f"^{DOWN_NAME}",
    tuple(CTRL_LEFT): f"^{LEFT_NAME}",
    tuple(CTRL_RIGHT): f"^{RIGHT_NAME}",
    tuple(CTRL__): "Esc",
    tuple(ALT_UP): f"{ALT_NAME}{UP_NAME}",
    tuple(ALT_DOWN): f"{ALT_NAME}{DOWN_NAME}",
    tuple(ALT_LEFT): f"{ALT_NAME}{LEFT_NAME}",
    tuple(ALT_RIGHT): f"{ALT_NAME}{RIGHT_NAME}",
    tuple(ALT_J): f"{ALT_NAME}j",
    tuple(ALT_K): f"{ALT_NAME}k",
}

NORMAL = "normal"
SCROLL = "scroll"
INPUT = "input"
GLOBAL = "global"
HELP = "help"

bindings = {
    NORMAL: {},
    SCROLL: {},
    INPUT: {},
    GLOBAL: {},
    HELP: {},
}

descriptions = {
    HELP: {},
    NORMAL: {},
    SCROLL: {},
    INPUT: {},
    GLOBAL: {},
}


def generic_key_to_name(key):
    name = chr(key)
    if not name.isprintable():
        name = f"[{ord(name)}]"
    return name


def seq_to_name(seq, fallback=True):
    if not isinstance(seq, tuple):
        seq = tuple(seq)
    name = None
    if seq in CTRL_TO_NAME:
        name = f"^{CTRL_TO_NAME[seq]}"
    elif seq in SEQ_TO_NAME:
        name = SEQ_TO_NAME[seq]
    elif len(seq) == 1:
        key = seq[0]
        name = generic_key_to_name(key)
    elif fallback:
        name = "".join(generic_key_to_name(o) for o in seq)
    return name


def is_multi(k):
    return isinstance(k, list) and isinstance(k[0], (list, tuple, str))


def key_to_seq(k):
    if isinstance(k, str):
        return tuple(ord(c) for c in k)
    elif is_multi(k):
        result = []
        for e in k:
            result.extend(key_to_seq(e))
        return tuple(result)
    elif isinstance(k, (list, tuple)):
        return tuple(k)
    else:
        raise RuntimeError(k)


def bind(mode, *keys, description=None, custom_bindings=None, custom_descriptions=None):
    used_bindings = custom_bindings if custom_bindings is not None else bindings
    used_descriptions = custom_descriptions if custom_descriptions is not None else descriptions

    def wrapper(fn):
        description_keys = []
        for key in keys:
            if not is_multi(key):
                key = [key]
            sequences = [key_to_seq(k) for k in key]
            name = "".join(seq_to_name(seq) for seq in sequences)
            description_keys.append(name)
        description_keys = tuple(description_keys)
        used_descriptions[mode][description_keys] = description or fn.__name__
        for k in keys:
            used_bindings[mode][key_to_seq(k)] = fn
        return fn

    return wrapper



import asyncio

from multiplex.actions import SetTitle, ToggleCollapse, ToggleWrap
from multiplex.refs import STOP


class Controller:
    def __init__(self, title, thread_safe=False):
        self.title = title
        self.queue: asyncio.Queue = None
        self.thread_safe = thread_safe
        self._loop: asyncio.AbstractEventLoop = None
        self._pre_init_queue = []

    def _init(self):
        self.queue = asyncio.Queue()
        self._loop = asyncio.get_event_loop()
        for data in self._pre_init_queue:
            self.write(data)

    def write(self, data):
        if not self.queue:
            self._pre_init_queue.append(data)
            return
        if self.thread_safe:
            self._loop.call_soon_threadsafe(self.queue.put_nowait, data)
        else:
            self.queue.put_nowait(data)

    def set_title(self, title):
        self.write(SetTitle(title))

    def collapse(self):
        self.write(ToggleCollapse(True))

    def expand(self):
        self.write(ToggleCollapse(False))

    def toggle_collapse(self):
        self.write(ToggleCollapse())

    def wrap(self):
        self.write(ToggleWrap(True))

    def nowrap(self):
        self.write(ToggleWrap(False))

    def toggle_wrap(self):
        self.write(ToggleWrap())

    def done(self):
        self.write(STOP)



import asyncio
import os
import sys
from itertools import cycle

import click
from multiplex.exceptions import IPCException
from multiplex.ipc import Client, get_env_stream_id
from multiplex.iterator import MULTIPLEX_SOCKET_PATH
from multiplex.multiplex import Multiplex


async def ipc_mode(socket_path, process, title, box_height, wait, load):
    client = Client(socket_path)
    first = process[0] if not load else None
    env = os.environ.copy()
    cwd = os.getcwd()
    if load:
        await client.load(load)
    elif first == "@":
        stream_id = get_env_stream_id()
        title = title[0] or " ".join(process[1:])
        await client.split(title, box_height[0], stream_id)
    elif first == "@save":
        await client.save()
    elif first == ":":
        await client.quit()
    elif first in {"/", "+", "-"}:
        value = None if first == "/" else True if first == "-" else False
        await client.toggle_collapse(value)
    elif len(process) == 1:
        await client.add(first, title[0], box_height[0], wait[0], cwd, env)
    else:
        actions = []
        for p, t, h, w in zip(process, cycle(title), cycle(box_height), cycle(wait)):
            actions.append(client.add_request_body(p, title=t, box_height=h, wait=w, cwd=cwd, env=env))
        await client.batch(actions)


def direct_mode(process, title, verbose, box_height, auto_collapse, output_path, load, socket_path, buffer_lines):
    multiplex = Multiplex(
        verbose=verbose,
        box_height=box_height[0],
        auto_collapse=auto_collapse,
        output_path=output_path,
        socket_path=socket_path,
        buffer_lines=buffer_lines,
    )
    for p, t, h in zip(process, cycle(title), cycle(box_height)):
        multiplex.add(p, title=t, box_height=h)
    multiplex.run(load)
    for iterator in multiplex.viewer.iterators:
        exit_code = iterator.metadata.get("exit_code")
        if exit_code:
            sys.exit(exit_code)


def validate(box_height, process, socket_path, title, wait, load):
    if load:
        if not os.path.isdir(load):
            raise click.ClickException(f"No such dir: {load}")
        if not os.path.exists(os.path.join(load, "metadata.json")):
            raise click.ClickException(f"Invalid directory: {load}")
        return
    if not process:
        raise click.ClickException("At least one command is required")
    if title and len(title) != len(process):
        raise click.ClickException(
            f"Each process should have a title, but {len(title)} titles and {len(process)} processes were supplied"
        )
    if wait and not socket_path:
        raise click.ClickException(f"wait can only be used when running inside a process started by multiplex")
    if len(wait) > 1 and len(wait) != len(process):
        raise click.ClickException(
            f"wait should be supplied either once to apply to all boxes or once for each process, "
            f"but {len(wait)} wait's and {len(process)} processes were supplied"
        )
    if len(box_height) > 1 and len(box_height) != len(process):
        raise click.ClickException(
            f"Box height should be supplied either once to apply to all boxes or once for each process, "
            f"but {len(box_height)} box heights and {len(process)} processes were supplied"
        )


@click.command()
@click.argument("process", nargs=-1)
@click.option("-t", "--title", multiple=True)
@click.option("-b", "--box-height", type=int, multiple=True)
@click.option("--wait/--no-wait", "-w/-W", multiple=True, default=None)
@click.option("-l", "--load")
@click.option(
    "--buffer-lines",
    type=int,
    envvar="MULTIPLEX_BUFFER_LINES",
    help="By default, buffer length is unbounded. Use this to have a maximum number of lines for each " "buffer.",
)
@click.option(
    "-a/-A",
    "--auto-collapse/--no-auto-collapse",
    type=bool,
    envvar="MULTIPLEX_AUTO_COLLAPSE",
    help="Collapse buffers automatically upon successful completion",
)
@click.option(
    "-o",
    "--output-path",
    help="Root directory to use when saving output",
    default=os.getcwd(),
    envvar="MULTIPLEX_OUTPUT_PATH",
)
@click.option(
    "-s",
    "--socket-path",
    envvar=MULTIPLEX_SOCKET_PATH,
)
@click.option(
    "--server",
    is_flag=True,
    help="This should only be used when socket path is provided explicity to instantiate the server. "
    "Otherwise, the command is assumed to be executed in IPC mode",
)
@click.help_option("-h", "--help")
@click.version_option(None, "--version")
@click.option("-v", "--verbose", is_flag=True)
def main(
    process, title, verbose, box_height, auto_collapse, output_path, wait, load, socket_path, buffer_lines, server
):
    validate(
        box_height=box_height,
        process=process,
        socket_path=socket_path,
        title=title,
        wait=wait,
        load=load,
    )

    title = title or [None]
    box_height = box_height or [None]
    wait = wait or [True]

    if socket_path and not server:
        try:
            asyncio.run(
                ipc_mode(
                    socket_path=socket_path,
                    process=process,
                    title=title,
                    box_height=box_height,
                    wait=wait,
                    load=load,
                )
            )
        except IPCException as e:
            raise click.ClickException(str(e))
    else:
        direct_mode(
            process=process,
            title=title,
            verbose=verbose,
            box_height=box_height,
            auto_collapse=auto_collapse,
            output_path=output_path,
            load=load,
            socket_path=socket_path,
            buffer_lines=buffer_lines,
        )


if __name__ == "__main__":
    main()



import argparse
import asyncio
import io
import random
import threading
import time

import colors
from colors.colors import _color_code as cc

from multiplex import ansi, Process
from multiplex import Multiplex, Controller
from multiplex.log import init_logging


def run_simple():
    async def text_generator(index):
        for i in range(num_iterations):
            output = f"iterator-number-#{index + 1}-({i + 1})|" * 7
            output = f"{output}\n"
            yield output
            await asyncio.sleep(random.random() / 10)
            await asyncio.sleep(0.5)
        yield "done"

    num_iterations = 2000
    num_iterators = 20
    iterators = [(text_generator(i), f"It #{i + 1}") for i in range(num_iterators)]
    return iterators


def run_colors():
    async def text_generator():
        for i in range(num_iterations):
            r = random.randint(0, 256)
            g = random.randint(0, 256)
            b = random.randint(0, 256)
            output = "".join(f'{colors.color(f"hello-{i}-", (r, g, b))}' for i in range(30))
            output = f"{output}\n"
            yield output
            await asyncio.sleep(random.random() / 10)
        yield "done"

    num_iterations = 2000
    num_iterators = 3
    iterators = [(text_generator, f"It #{i + 1}") for i in range(num_iterators)]
    return iterators


def run_dynamic():
    async def text_generator():
        for i in range(num_iterations):
            for j in range(10000):
                yield "hello"
                await asyncio.sleep(0.1)
                yield "\rgoodbye"
                await asyncio.sleep(0.1)
                yield "\r" + ansi.CLEAR_LINE
            yield "\n"
            await asyncio.sleep(random.random() / 10)
            await asyncio.sleep(0.5)
        yield "done"

    num_iterations = 2000
    num_iterators = 3
    iterators = [(text_generator, f"It #{i + 1}") for i in range(num_iterators)]
    return iterators


def run_style():
    async def text_generator():
        for i in range(num_iterations):

            fr = random.randint(0, 256)
            fg = random.randint(0, 256)
            fb = random.randint(0, 256)

            br = random.randint(0, 256)
            bg = random.randint(0, 256)
            bb = random.randint(0, 256)

            def code(*codes):
                return f'{ansi.CSI}{";".join(str(c) for c in codes)}m'

            reset = code(0)

            text_buffer = io.StringIO()
            for j in range(100):
                text_buffer.write(code(cc("red", 30)))
                text_buffer.write(code(cc("green", 40)))
                text_buffer.write(f"some text {j} ")
                text_buffer.write(code(cc((fr, fg, fb), 30)))
                text_buffer.write(code(cc((br, bg, bb), 40)))
                text_buffer.write(f"some text {j} ")
                text_buffer.write(code(3, 4, 7))
                text_buffer.write(f"some text {j} ")
                text_buffer.write(code(24, 9, 1))
                text_buffer.write(f"some text {j} ")
                text_buffer.write(reset)
            output = text_buffer.getvalue()
            output = f"{output}\n"
            yield output
            await asyncio.sleep(random.random() / 10)
            await asyncio.sleep(0.5)
        yield "done"

    num_iterations = 2000
    num_iterators = 3
    iterators = [(text_generator, f"It #{i + 1}") for i in range(num_iterators)]
    return iterators


def run_processes():
    return ["gls -la --group-directories-first --color=always"]


def run_process_desc():
    return [
        Process("gls"),
        Process("gls -la --group-directories-first --color=always".split(" ")),
        Process("gls -la".split(" ")),
    ]


def run_controller():
    multplex = Multiplex()
    c1 = Controller("runner1")
    c2 = Controller("runner2")
    multplex.add(c1)
    multplex.add(c2)

    async def runner(c):
        c.write("some data 1\n")
        await asyncio.sleep(1)
        c.write("some data 2\n")
        await asyncio.sleep(1)
        c.write("some data 2\n")
        await asyncio.sleep(1)
        c.write("some data 2\n")
        c.set_title(f"{c.title} [done]")
        c.collapse()

    run(multplex, runner(c1), runner(c2))


def run_controller_thread_safe():
    multiplex = Multiplex()
    c1 = Controller("runner1", thread_safe=True)
    c2 = Controller("runner2", thread_safe=True)
    multiplex.add(c1)
    multiplex.add(c2)

    def runner(c):
        c.write("some data 1\n")
        time.sleep(1)
        c.write("some data 2\n")
        time.sleep(1)
        c.write("some data 2\n")
        time.sleep(1)
        c.write("some data 2\n")
        c.set_title(f"{c.title} [done]")
        c.collapse()

    threads = [threading.Thread(target=runner, args=(c,)) for c in [c1, c2]]
    for t in threads:
        t.daemon = True
        t.start()
    multiplex.run()


def run_live():
    obj = "echo $RANDOM; sleep 5; echo $RANDOM"

    multi = Multiplex(box_height=3)

    async def runner():
        while not multi.viewer or not multi.viewer.stopped:
            multi.add(obj)
            await asyncio.sleep(0.1)

    run(multi, runner())


def run_live_thread_safe():
    obj = "echo $RANDOM; sleep 5; echo $RANDOM"

    multi = Multiplex(box_height=3)

    def runner(_):
        while not multi.viewer or not multi.viewer.stopped:
            multi.add_thread_safe(obj)
            time.sleep(1)

    threads = [threading.Thread(target=runner, args=(c,)) for c in [1]]
    for t in threads:
        t.daemon = True
        t.start()
    multi.run()


def run_multiline():
    from easyansi import cursor, screen

    m = Multiplex(verbose=True)

    async def lines():
        yield "one\n"
        yield "two\n"
        await asyncio.sleep(0.3)
        yield "three"
        yield screen.clear_line_code(2)
        yield "ten"
        yield cursor.up_code(1)
        yield cursor.locate_column_code(4)
        yield "new two"
        yield cursor.up_code(1)
        yield cursor.locate_column_code(0)
        yield "n\n"
        yield "b\n"
        yield "c\n"
        yield "r\r\r\r\r\r\n"

    m.add(lines)
    m.run()


whats = {
    "1": run_simple,
    "2": run_processes,
    "3": run_colors,
    "4": run_dynamic,
    "5": run_style,
    "6": run_controller,
    "7": run_controller_thread_safe,
    "8": run_live,
    "9": run_live_thread_safe,
    "10": run_multiline,
    "11": run_process_desc,
}


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("what")
    return parser.parse_args()


def run(multi, *other):
    async def cor():
        return await asyncio.gather(multi.run_async(), *other)

    try:
        asyncio.run(cor())
    except KeyboardInterrupt:
        pass
    finally:
        multi.cleanup()


def main():
    init_logging()
    fn = whats.get(parse_args().what, run_simple)
    result = fn()
    if isinstance(result, list):
        multi = Multiplex(verbose=True)
        for i in result:
            title = None
            if isinstance(i, tuple):
                i, title = i
            multi.add(i, title)
        multi.run()


if __name__ == "__main__":
    main()



import asyncio

from aiostream import streamcontext
from aiostream.test_utils import assert_aiter
import pytest

from multiplex import iterator as _iterator
from multiplex.controller import Controller
from multiplex.iterator import to_iterator, STOP

pytestmark = pytest.mark.asyncio

ACTION = object()


async def collect(source):
    result = []
    async with streamcontext(source) as streamer:
        async for item in streamer:
            result.append(item)
    return result


@pytest.fixture
def patch_actions(monkeypatch):
    monkeypatch.setattr(_iterator, "BoxActions", lambda *_, **__: ACTION)
    monkeypatch.setattr(_iterator, "UpdateMetadata", lambda *_, **__: ACTION)


async def test_async_generator_input():
    async def g():
        for i in range(1):
            yield i

    iterator = await to_iterator(g())
    assert iterator.title == "g"
    assert iterator.inner_type == "async_generator"
    await assert_aiter(iterator.iterator, [0])


async def test_generator_input():
    def g():
        for i in range(1):
            yield i

    iterator = await to_iterator(g())
    assert iterator.title == "g"
    assert iterator.inner_type == "generator"
    await assert_aiter(iterator.iterator, [0])


async def test_async_process_pipe_stdout(patch_actions):
    p = await asyncio.subprocess.create_subprocess_shell(
        cmd="printf 1; sleep 0.01; printf 2",
        stdout=asyncio.subprocess.PIPE,
    )
    iterator = await to_iterator(p)
    assert iterator.title == "Process"
    assert iterator.inner_type == "async_process"
    await assert_aiter(iterator.iterator, ["1", "2", ACTION])


async def test_async_process_pipe_stderr(patch_actions):
    p = await asyncio.subprocess.create_subprocess_shell(
        cmd="printf 1 1>&2",
        stderr=asyncio.subprocess.PIPE,
    )
    iterator = await to_iterator(p)
    assert iterator.title == "Process"
    assert iterator.inner_type == "async_process"
    await assert_aiter(iterator.iterator, ["1", ACTION])


async def test_async_process_pipe_stdout_stederr(patch_actions):
    p = await asyncio.subprocess.create_subprocess_shell(
        cmd="""
            printf 1 1>&2 && \
            sleep 0.01 && \
            printf 2 && \
            sleep 0.01 && \
            printf 3 1>&2 && \
            sleep 0.01 && \
            printf 4 
        """,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    iterator = await to_iterator(p)
    assert iterator.title == "Process"
    assert iterator.inner_type == "async_process"
    await assert_aiter(iterator.iterator, ["1", "2", "3", "4", ACTION])


async def test_async_process_pipe_backslash_r(patch_actions):
    p = await asyncio.subprocess.create_subprocess_shell(
        cmd=f"printf hello; sleep 0.01; printf \r; sleep 0.01; echo goodbye",
        stdout=asyncio.subprocess.PIPE,
    )
    iterator = await to_iterator(p)
    assert iterator.title == "Process"
    assert iterator.inner_type == "async_process"
    await assert_aiter(
        iterator.iterator,
        [
            "hello",
            "\r",
            "goodbye\n",
            ACTION,
        ],
    )


async def test_async_stream_reader():
    p = await asyncio.subprocess.create_subprocess_shell(
        cmd="printf 1; sleep 0.01; printf 2",
        stdout=asyncio.subprocess.PIPE,
    )
    iterator = await to_iterator(p.stdout)
    assert iterator.title == "StreamReader"
    assert iterator.inner_type == "stream_reader"
    await assert_aiter(iterator.iterator, ["1", "2"])


async def test_async_iter():
    class It:
        def __init__(self, values):
            self.current_index = 0
            self.values = values

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.current_index < len(self.values):
                index = self.current_index
                self.current_index += 1
                return self.values[index]
            raise StopAsyncIteration

    vals = [1, 2, 3]
    iterator = await to_iterator(It(vals))
    assert iterator.title == "It"
    assert iterator.inner_type == "async_iterable"
    await assert_aiter(iterator.iterator, vals)


async def test_iterable():
    class It:
        def __init__(self, values):
            self.current_index = 0
            self.values = values

        def __iter__(self):
            return self

        def __next__(self):
            if self.current_index < len(self.values):
                index = self.current_index
                self.current_index += 1
                return self.values[index]
            raise StopIteration

    vals = [1, 2, 3]
    iterator = await to_iterator(It(vals))
    assert iterator.title == "It"
    assert iterator.inner_type == "iterable"
    await assert_aiter(iterator.iterator, vals)


async def test_coroutine(patch_actions):
    c = asyncio.subprocess.create_subprocess_shell(
        cmd="printf hello",
        stdout=asyncio.subprocess.PIPE,
    )
    iterator = await to_iterator(c)
    assert iterator.title == "Process"
    assert iterator.inner_type == "async_process"
    await assert_aiter(iterator.iterator, ["hello", ACTION])


async def test_function():
    async def fn():
        yield "1"

    iterator = await to_iterator(fn)
    assert iterator.inner_type == "async_generator"
    assert iterator.title == "fn"
    await assert_aiter(iterator.iterator, ["1"])


async def test_method():
    class C:
        val1 = "1"
        val2 = "2"

        async def fn1(self):
            yield self.val1

        @classmethod
        async def fn2(cls):
            yield cls.val2

        @staticmethod
        async def fn3():
            yield "3"

    c = C()
    iterator = await to_iterator(c.fn1)
    assert iterator.title == "fn1"
    assert iterator.inner_type == "async_generator"
    await assert_aiter(iterator.iterator, ["1"])

    iterator = await to_iterator(c.fn2)
    assert iterator.title == "fn2"
    assert iterator.inner_type == "async_generator"
    await assert_aiter(iterator.iterator, ["2"])

    iterator = await to_iterator(c.fn3)
    assert iterator.title == "fn3"
    assert iterator.inner_type == "async_generator"
    await assert_aiter(iterator.iterator, ["3"])


async def test_path(tmp_path):
    path = tmp_path / "mock.txt"
    text = "hello\ngoodbye"
    path.write_text(text)
    iterator = await to_iterator(path)
    assert iterator.inner_type == "path"
    assert iterator.title == str(path)
    await assert_aiter(iterator.iterator, [text])


async def test_str_process(patch_actions):
    cmd = "printf hello"
    iterator = await to_iterator(cmd)
    assert iterator.inner_type == "async_process"
    assert iterator.title == cmd
    await assert_aiter(iterator.iterator, [ACTION, "hello", None, ACTION])


async def test_str_path(tmp_path):
    path = tmp_path / "mock.txt"
    text = "hello\ngoodbye"
    path.write_text(text)
    iterator = await to_iterator(f"file://{path}")
    assert iterator.inner_type == "path"
    assert iterator.title == str(path)
    await assert_aiter(iterator.iterator, [text])


async def test_setsize(patch_actions, monkeypatch):
    cols = 100
    rows = 50
    monkeypatch.setenv("COLUMNS", str(cols))
    monkeypatch.setenv("LINES", str(rows))

    cmd = "stty size"
    iterator = await to_iterator(cmd)
    assert iterator.inner_type == "async_process"
    assert iterator.title == cmd
    items = await collect(iterator.iterator)
    assert items[1].strip() == f"{rows} {cols}"


async def test_controller():
    value = "data1"
    title = "title1"
    c = Controller(title)
    c.write(value)
    c.write(STOP)
    iterator = await to_iterator(c)
    assert iterator.inner_type == "controller"
    assert iterator.title == title
    await assert_aiter(iterator.iterator, [value])


async def test_controller_thead_safe():
    value = "data1"
    title = "title1"
    c = Controller(title, thread_safe=True)
    c.write(value)
    c.write(STOP)
    iterator = await to_iterator(c)
    assert iterator.inner_type == "controller"
    assert iterator.title == title
    await assert_aiter(iterator.iterator, [value])



import asyncio
import os
import subprocess

import pytest
from multiplex import Multiplex, to_iterator, Controller
from multiplex.ipc import Client

pytestmark = pytest.mark.asyncio


def test_exports():
    assert Controller
    assert Multiplex
    assert to_iterator


@pytest.mark.skipif("os.environ.get('MULTIPLEX_SOCKET_PATH')", reason="Running in demo")
async def test_sanity(tmpdir):
    output_dir = tmpdir / "output"
    socket_echo = tmpdir / "socket_loc"
    output_dir.mkdir()
    cmd = f"mp 'echo 1' 'echo 2' 'echo $MULTIPLEX_SOCKET_PATH > {socket_echo}' -o {output_dir}"
    master, slave = os.openpty()
    proc = await asyncio.subprocess.create_subprocess_shell(
        cmd,
        stdin=slave,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    await asyncio.sleep(0.5)
    socket_path = socket_echo.read_text("utf-8").strip()
    client = Client(socket_path)
    await client.save()
    await client.quit()
    await proc.wait()
    output_dir = output_dir.listdir()[0]
    expected = ["1", "2", ""]
    listing = output_dir.listdir(sort=True)
    assert len(listing) == len(expected) + 1
    for i, ex in enumerate(expected):
        assert listing[i].read_text("utf-8").strip() == ex



def obj(name):
    return type(name, (), {})()


REDRAW = obj("REDRAW")
RECALC = obj("RECALC")
SPLIT = obj("SPLIT")
STOP = obj("STOP")
QUIT = obj("QUIT")
ALL_DOWN = obj("ALL_DOWN")
SAVE = obj("SAVE")
OUTPUT_SAVED = obj("OUTPUT_SAVED")
STREAM_DONE = obj("STREM_DONE")



from multiplex.buffer import Buffer


def test_buffer_kitchen():
    width = 5
    buffer = Buffer(width)
    assert buffer.width == width
    text = "123456789"
    buffer.write(text)
    assert buffer.get_max_line(wrap=False) == 0
    assert buffer.get_max_line(wrap=True) == 1
    assert buffer.get_lines(1, 0, width, 0, wrap=False) == [(9, "12345")]
    assert buffer.get_lines(1, 0, width, 0, wrap=True) == [(5, "12345")]
    assert buffer.get_lines(2, 0, width, 0, wrap=False) == [(9, "12345"), (0, " " * 5)]
    assert buffer.get_lines(2, 0, width, 0, wrap=True) == [(5, "12345"), (4, "6789 ")]
    assert buffer.get_lines(1, 1, width, 0, wrap=True) == [(4, "6789 ")]
    assert buffer.get_lines(1, 0, width, 3, wrap=False) == [(9, "45678")]

    width = 3
    buffer.width = width
    assert buffer.get_max_line(wrap=False) == 0
    assert buffer.get_max_line(wrap=True) == 2
    assert buffer.get_lines(1, 0, width, 0, wrap=False) == [(9, "123")]
    assert buffer.get_lines(3, 0, width, 0, wrap=True) == [(3, "123"), (3, "456"), (3, "789")]

    assert buffer.raw_buffer.getvalue() == text

    buffer.write("\nabcd")
    assert buffer.convert_line_number(0, from_wrapped=False) == 0
    assert buffer.convert_line_number(1, from_wrapped=False) == 3
    assert buffer.convert_line_number(0, from_wrapped=True) == 0
    assert buffer.convert_line_number(2, from_wrapped=True) == 0
    assert buffer.convert_line_number(3, from_wrapped=True) == 1
    assert buffer.convert_line_number(4, from_wrapped=True) == 1



import collections
import io
import shutil

import pyte
from pyte import graphics as g
from pyte.screens import Char, wcwidth, Margins

from multiplex.ansi import CSI

TERMINATE = "m"

UNDEFINED = object()
RESET_TEXT_ATTRS = set(list(range(1, 10)))
BOLD = 1

empty_meta = Char(
    None,
    fg=None,
    bg=None,
    bold=(),
    italics=None,
    underscore=None,
    strikethrough=None,
    reverse=None,
)

reset = f"{CSI}0{TERMINATE}"
index_to_char_meta = {0: empty_meta}
char_meta_to_index = {empty_meta: 0}
index_to_ansi = {0: reset}

counter = 0


class Screen(pyte.Screen):
    def __init__(self, columns, lines, line_buffer):
        super().__init__(columns, lines)
        self.line_buffer = line_buffer

    def reset(self):
        original_columns = self.columns
        original_lines = self.lines
        self.columns = 1
        self.lines = 1
        super().reset()
        self.cursor.attrs = self.default_char
        self.dirty.clear()
        self.columns = original_columns
        self.lines = original_lines
        self.tabstops = set(range(8, self.columns, 8))

    @property
    def default_char(self):
        return Char(data=" ", fg=0)

    def select_graphic_rendition(self, *attrs):
        if not attrs or attrs == (0,):
            self.cursor.attrs = self.default_char
            return

        fg = UNDEFINED
        bg = UNDEFINED
        added_text_attrs = set()
        removed_text_attrs = set()

        attrs = list(reversed(attrs))
        while attrs:
            attr = attrs.pop()
            if attr == 0:
                fg = None
                bg = None
                removed_text_attrs = RESET_TEXT_ATTRS
            elif attr in g.FG_ANSI:
                fg = (attr,)
            elif attr in g.BG:
                bg = (attr,)
            elif attr in g.FG_AIXTERM:
                fg = (attr,)
                added_text_attrs.add(BOLD)
            elif attr in g.BG_AIXTERM:
                bg = (attr,)
                added_text_attrs.add(BOLD)
            elif attr in (g.FG_256, g.BG_256):
                n = attrs.pop()
                if n == 5:
                    value = attr, n, attrs.pop()
                    if attr == g.FG_256:
                        fg = value
                    else:
                        bg = value
                elif n == 2:
                    value = attr, n, attrs.pop(), attrs.pop(), attrs.pop()
                    if attr == g.FG_256:
                        fg = value
                    else:
                        bg = value
            elif 1 <= attr <= 9:
                added_text_attrs.add(attr)
            elif 21 <= attr <= 29:
                removed_text_attrs.add(attr)

        current_meta = index_to_char_meta[self.cursor.attrs.fg]
        current_text_attrs = set(current_meta.bold)
        new_text_attrs = (current_text_attrs | added_text_attrs) - removed_text_attrs

        replace = {}
        if fg is not UNDEFINED:
            replace["fg"] = fg
        if bg is not UNDEFINED:
            replace["bg"] = bg
        replace["bold"] = tuple(sorted(new_text_attrs))
        new_char_meta = current_meta._replace(**replace)

        if new_char_meta in char_meta_to_index:
            index = char_meta_to_index[new_char_meta]
        else:
            global counter
            counter += 1
            index = counter
            char_meta_to_index[new_char_meta] = index
            index_to_char_meta[index] = new_char_meta
            codes = []
            c = new_char_meta
            if c.fg:
                codes.extend(c.fg)
            if c.bg:
                codes.extend(c.bg)
            if c.bold:
                codes.extend(list(c.bold))
            ansi = f'{CSI}{";".join(str(c) for c in codes)}{TERMINATE}'
            index_to_ansi[index] = ansi

        self.cursor.attrs = Char(" ", fg=index)

    def erase_in_display(self, how=0, private=False):
        interval = None
        if how == 0:
            interval = range(self.cursor.y + 1, self.line_buffer.max_lines + 1)
        elif how == 1:
            interval = range(self.cursor.y)
        elif how == 2 or how == 3:
            interval = range(self.line_buffer.min_line, self.line_buffer.max_lines + 1)
        self.dirty.update(interval)
        for y in interval:
            line = self.buffer[y]
            for x in line:
                line[x] = self.cursor.attrs
        if how == 0 or how == 1:
            self.erase_in_line(how)

    def erase_in_line(self, how=0, private=False):
        self.dirty.add(self.cursor.y)
        line = self.buffer[self.cursor.y]
        keys = line.keys()
        columns = (max(keys) if keys else 0) + 1
        interval = None
        if how == 0:
            interval = range(self.cursor.x, columns)
        elif how == 1:
            interval = range(self.cursor.x + 1)
        elif how == 2:
            interval = range(columns)
        line = self.buffer[self.cursor.y]
        for x in interval:
            line[x] = self.cursor.attrs

    def index(self):
        self.cursor_down()

    def reverse_index(self):
        max_line = self.line_buffer.max_line
        min_line = self.line_buffer.min_line
        top, bottom = self.margins or Margins(min_line, max_line)
        if self.cursor.y == top:
            self.dirty.update(range(min_line, max_line + 1))
            for y in range(bottom, top, -1):
                self.buffer[y] = self.buffer[y - 1]
            self.buffer.pop(top, None)
        else:
            self.cursor_up()

    @property
    def display(self):
        # overriding this so we don't evaluate all virtual lines/columns
        # when debugging, etc...
        return []


class LinedBuffer:
    BIG = 1000000

    def __init__(self, width=None):
        self.width = width or self.BIG
        self.screen = Screen(lines=self.BIG, columns=self.width, line_buffer=self)
        self.stream = pyte.Stream(screen=self.screen)
        self.max_line = 0
        self.min_line = 0
        self.raw_to_self = {}
        self.self_to_raw = {}

    def write(self, data):
        self.stream.feed(data)
        return self._update()

    def _update(self):
        dirty = self.screen.dirty
        if dirty:
            dirty_list = sorted(list(dirty))
            self.max_line = max(self.max_line, *dirty_list)
            self.screen.dirty.clear()
            return dirty_list
        return []

    def remove_lines(self, lines, start_line):
        new_min_line = start_line + lines
        for i in range(start_line, new_min_line):
            self.screen.buffer.pop(i, None)
        return new_min_line

    def get_lines(self, lines, start_line, columns, start_column):
        result = []
        if start_line > self.max_line:
            return result
        last_char_meta_index = 0
        buffer = self.screen.buffer
        for line_num in range(start_line, lines + start_line):
            screen_line = buffer[line_num]
            keys = screen_line.keys()
            line_length = (max(keys) + 1) if keys else 0
            is_wide_char = False
            current_line_buffer = io.StringIO()
            for x in range(start_column, columns + start_column):
                if is_wide_char:  # Skip stub
                    is_wide_char = False
                    continue
                current_char = screen_line[x]
                char_meta_index = current_char.fg
                if char_meta_index != last_char_meta_index:
                    current_line_buffer.write(reset)
                    if char_meta_index:
                        current_line_buffer.write(index_to_ansi[char_meta_index])
                    last_char_meta_index = char_meta_index
                char_data = current_char.data
                current_line_buffer.write(char_data)
                assert sum(map(wcwidth, char_data[1:])) == 0
                is_wide_char = wcwidth(char_data[0]) == 2
            # add reset at the end
            if last_char_meta_index and line_num == lines + start_line - 1:
                current_line_buffer.write(reset)
            result.append((line_length, current_line_buffer.getvalue()))
        return result


class CappedRawBuffer:
    def __init__(self, buffer_lines):
        self._deque = collections.deque(maxlen=buffer_lines + 1)

    def write(self, data):
        if self._deque:
            data = self._deque.pop() + data
        self._deque.append(data)

    def writeline(self, line):
        self._deque.append(line)

    def newline(self):
        pass

    def getvalue(self):
        return "\n".join(self._deque)


class UncappedRawBuffer:
    def __init__(self):
        self._io = io.StringIO()

    def write(self, data):
        self._io.write(data)

    def writeline(self, line):
        self._io.write(line)

    def newline(self):
        self._io.write("\n")

    def getvalue(self):
        return self._io.getvalue()


class Buffer:
    def __init__(self, width=None, buffer_lines=None):
        self.buffer_lines = buffer_lines
        self.raw_buffer = CappedRawBuffer(buffer_lines) if buffer_lines else UncappedRawBuffer()
        self.raw_lines = 0
        self.lined_buffer = LinedBuffer()
        self.wrapping_buffer = self._new_wrapping_buffer(width)

    @staticmethod
    def _new_wrapping_buffer(width):
        return LinedBuffer(width or shutil.get_terminal_size().columns)

    def get_lines(self, lines, start_line, columns, start_column, wrap):
        return self._get_buffer(wrap).get_lines(
            lines=lines,
            start_line=start_line,
            columns=columns,
            start_column=start_column,
        )

    def write(self, data, buffers=None, skip_raw=False):
        buffers = buffers or (self.lined_buffer, self.wrapping_buffer)
        lines = data.split("\n")
        for i, line in enumerate(lines):
            if not skip_raw:
                if i:
                    self.raw_lines += 1
                    self.raw_buffer.writeline(line)
                else:
                    self.raw_buffer.write(line)
                if i < len(lines) - 1:
                    self.raw_buffer.newline()
                current_raw_line = self.raw_lines
            else:
                current_raw_line = self.raw_lines - len(lines) + i + 1
            if i < len(lines) - 1:
                maybe_slash_r = "" if line and line[-1] == "\r" else "\r"
                line = f"{line}{maybe_slash_r}\n"
            for buffer in buffers:
                dirty_lines = buffer.write(line)
                if dirty_lines:
                    buffer.raw_to_self[current_raw_line] = dirty_lines[0]
                    for dl in dirty_lines:
                        buffer.self_to_raw[dl] = current_raw_line
            if not skip_raw and self.buffer_lines:
                lined_buffer = self.lined_buffer
                wrapping_buffer = self.wrapping_buffer
                total_lines = lined_buffer.max_line - lined_buffer.min_line + 1
                if total_lines > self.buffer_lines:
                    remove_lined = total_lines - self.buffer_lines
                    lined_buffer.min_line = lined_buffer.remove_lines(remove_lined, lined_buffer.min_line)
                    remove_wrapping = (
                        self.convert_line_number(lined_buffer.min_line, fail_on_error=True) - wrapping_buffer.min_line
                    )
                    wrapping_buffer.min_line = wrapping_buffer.remove_lines(remove_wrapping, wrapping_buffer.min_line)

    @property
    def width(self):
        return self.wrapping_buffer.width

    @width.setter
    def width(self, value):
        self.wrapping_buffer = self._new_wrapping_buffer(value)
        self.write(self.raw_buffer.getvalue(), skip_raw=True, buffers=[self.wrapping_buffer])

    def get_min_line(self, wrap):
        return self._get_buffer(wrap).min_line

    def get_max_line(self, wrap):
        return self._get_buffer(wrap).max_line

    def get_cursor(self, wrap):
        cursor = self._get_buffer(wrap).screen.cursor
        return cursor.x, cursor.y

    def _get_buffer(self, wrap):
        return self.wrapping_buffer if wrap else self.lined_buffer

    def convert_line_number(self, line_number, from_wrapped=False, fail_on_error=False):
        from_buffer = self.wrapping_buffer if from_wrapped else self.lined_buffer
        to_buffer = self.lined_buffer if from_wrapped else self.wrapping_buffer
        raw_line_number = from_buffer.self_to_raw.get(line_number)
        if raw_line_number is None:
            if fail_on_error:
                raise RuntimeError(f"No raw line for line {line_number}")
            else:
                # TODO log warning
                return 0
        to_line_number = to_buffer.raw_to_self.get(raw_line_number)
        if to_line_number is None:
            if fail_on_error:
                raise RuntimeError(f"No to_line for raw_line {raw_line_number} [line={line_number}]")
            else:
                # TODO log warning
                return 0
        return to_line_number



import logging

from multiplex import ansi
from multiplex.buffer import Buffer
from multiplex.enums import ViewLocation

logger = logging.getLogger("multiplex.box")


class BoxHolder:
    def __init__(self, index, iterator, box_height, viewer):
        self.id = id(self)
        self.index = index
        self.iterator = iterator
        self.buffer = Buffer(buffer_lines=viewer.buffer_lines)
        self.state = BoxState(box_height)
        self.box = TextBox(viewer, self)


class BoxState:
    def __init__(self, box_height):
        self.wrap = True
        self.auto_scroll = True
        self.input_mode = False
        self.stream_done = False
        self.collapsed = False
        self.buffer_start_line = 0
        self.first_column = 0
        self.view_longest_line = 0
        self.text = None
        self.changed_height = box_height is not None
        self.box_height = box_height


class TextBox:
    def __init__(self, view, holder):
        self.view = view
        self.holder = holder
        self.buffer = self.holder.buffer
        self.state = self.holder.state

    @property
    def index(self):
        return self.holder.index

    def update(self):
        lines = self.update_text()

        screen_y_1, location_1 = self.view.get_box_top_line(self.index)
        screen_y_2, location_2 = self.view.get_box_bottom_line(self.index)

        if location_1 == ViewLocation.ABOVE:
            lines = lines[screen_y_1:]
            screen_y_1 = 0
        if location_2 == ViewLocation.BELOW:
            lines = lines[:-screen_y_2]
            screen_y_2 = self.view.get_max_box_line()

        logger.debug(
            f"{self.index}:"
            f"\t{screen_y_1}"
            f"\t{screen_y_2}"
            f"\t{location_1}"
            f"\t{location_2}"
            f"\t[{self.view.lines},{self.view.cols}]"
        )

        ansi.text_box(
            from_row=screen_y_1,
            to_row=screen_y_2,
            text="".join(lines),
        )

    def update_text(self):
        if self.state.auto_scroll and not self.state.stream_done:
            self.state.buffer_start_line = self.max_start_line
        lines = self.buffer.get_lines(
            lines=self.num_view_lines,
            start_line=self.state.buffer_start_line,
            columns=self.view.cols,
            start_column=self.state.first_column,
            wrap=self.state.wrap,
        )
        if not self.state.wrap:
            value = max(line_length for line_length, _ in lines) if lines else 0
            self.state.view_longest_line = value
        return [line for _, line in lines]

    def move_line_up(self):
        return self.set_minmax_up_motion(self.state.buffer_start_line - 1)

    def move_line_down(self):
        return self.set_minmax_down_motion(self.state.buffer_start_line + 1)

    def move_page_up(self):
        return self.set_minmax_up_motion(self.state.buffer_start_line - self.num_view_lines)

    def move_page_down(self):
        return self.set_minmax_down_motion(self.state.buffer_start_line + self.num_view_lines)

    def move_half_page_up(self):
        return self.set_minmax_up_motion(self.state.buffer_start_line - self.num_view_lines // 2)

    def move_half_page_down(self):
        return self.set_minmax_down_motion(self.state.buffer_start_line + self.num_view_lines // 2)

    def move_all_up(self):
        self.state.buffer_start_line = self.min_start_line
        return self.index

    def move_all_down(self):
        self.state.buffer_start_line = self.max_start_line
        return self.index

    def move_right(self):
        state = self.state
        if state.wrap:
            return False
        state.first_column = min(state.first_column + 1, self.max_first_column)
        return self.index

    def move_left(self):
        state = self.state
        if state.wrap:
            return False
        state.first_column = max(0, state.first_column - 1)
        return self.index

    def move_half_screen_right(self):
        state = self.state
        if state.wrap:
            return False
        state.first_column = min(state.first_column + self.view.cols // 2, self.max_first_column)
        return self.index

    def move_half_screen_left(self):
        state = self.state
        if state.wrap:
            return False
        state.first_column = max(0, state.first_column - self.view.cols // 2)
        return self.index

    def move_right_until_end(self):
        state = self.state
        if state.wrap:
            return False
        state.first_column = self.max_first_column
        return self.index

    def move_left_until_start(self):
        state = self.state
        if state.wrap:
            return False
        state.first_column = 0
        return self.index

    def increase_box_height(self):
        self.state.box_height = min(self.view.get_max_box_line(), self.state.box_height + 1)
        self.state.changed_height = True
        return True

    def decrease_box_height(self):
        self.state.box_height = max(0, self.state.box_height - 1)
        self.state.changed_height = True
        return ansi.FULL_REFRESH

    def toggle_auto_scroll(self):
        self.state.auto_scroll = not self.state.auto_scroll
        return True

    def activate_input_mode(self):
        self.state.input_mode = True
        ansi.show_cursor()
        return True

    def exit_input_mode(self):
        self.state.input_mode = False
        ansi.hide_cursor()
        return True

    def toggle_wrap(self, value=None):
        initial_value = self.state.wrap
        new_value = value if value is not None else not initial_value
        if initial_value == new_value:
            return False
        self.state.first_column = 0
        self.state.wrap = new_value
        if not self.state.auto_scroll or self.state.stream_done:
            line = self.state.buffer_start_line
            result = self.buffer.convert_line_number(line, from_wrapped=initial_value)
            if result is None:
                logger.warning(f"No mathing line for conversion wrap: {new_value}, line: {line}")
                result = 0
            self.state.buffer_start_line = result
        return True

    def toggle_collapse(self, value=None):
        if value is not None:
            collapsed = value
        else:
            collapsed = not self.state.collapsed
        self.state.collapsed = collapsed
        return ansi.FULL_REFRESH if collapsed else True

    def strip_empty_lines(self, include_not_stream_done=False):
        if not include_not_stream_done and not self.state.stream_done:
            return
        self.state.box_height = min(self.state.box_height, self.num_buffer_lines)

    @property
    def num_view_lines(self):
        return self.view.lines - 1 if self.is_maximized else self.state.box_height

    @property
    def is_focused(self):
        return self.index == self.view.current_focused_box

    @property
    def is_visible(self):
        if self.view.maximized and not self.is_focused:
            return False
        if self.state.collapsed and not self.is_maximized:
            return False
        _, location = self.view.get_box_top_line(self.index)
        if location == ViewLocation.BELOW:
            return False
        _, location = self.view.get_box_bottom_line(self.index)
        if location == ViewLocation.ABOVE:
            return False
        return True

    @property
    def is_maximized(self):
        return self.view.maximized and self.is_focused

    @property
    def max_first_column(self):
        return max(0, self.state.view_longest_line - self.view.cols)

    @property
    def num_buffer_lines(self):
        buffer_min_line = self.buffer.get_min_line(self.state.wrap)
        buffer_max_line = self.buffer.get_max_line(self.state.wrap)
        return buffer_max_line - buffer_min_line + 1

    @property
    def min_start_line(self):
        return self.buffer.get_min_line(self.state.wrap)

    @property
    def max_start_line(self):
        buffer_min_line = self.buffer.get_min_line(self.state.wrap)
        buffer_max_line = self.buffer.get_max_line(self.state.wrap)
        return max(buffer_min_line, buffer_max_line - self.num_view_lines + 1)

    def set_minmax_down_motion(self, value):
        return self._set_min_max_motion(value, self.state.buffer_start_line)

    def set_minmax_up_motion(self, value):
        return self._set_min_max_motion(value, self.min_start_line)

    def set_width(self, width):
        raw_line = self.buffer.wrapping_buffer.self_to_raw.get(self.state.buffer_start_line, 0)
        self.buffer.width = width
        if self.state.wrap and (not self.state.auto_scroll or self.state.stream_done):
            self.state.buffer_start_line = self.buffer.wrapping_buffer.raw_to_self[raw_line]

    def _set_min_max_motion(self, value, min_value):
        self.state.buffer_start_line = max(min_value, min(self.max_start_line, value))
        return self.index



import asyncio
import functools
import os
import sys
import select
import termios
import tty

from multiplex import keys as _keys

initial_stdin_settings = None


def setup():
    global initial_stdin_settings
    initial_stdin_settings = termios.tcgetattr(sys.stdin)
    tty.setcbreak(sys.stdin.fileno())


def restore():
    termios.tcsetattr(sys.stdin, termios.TCSADRAIN, initial_stdin_settings)


def _has_data():
    return select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], [])


class InputReader:
    def __init__(self, viewer, bindings):
        self.viewer = viewer
        self.bindings = bindings
        self.pending = []

    async def read(self):
        while True:
            result = self._read_iteration()
            if result is not None:
                yield -1, result
            await asyncio.sleep(0.02)

    def _read_iteration(self):
        if not _has_data():
            return None
        skip_process = False
        is_input = self.viewer.is_input_mode
        keys = []
        while _has_data():
            key = os.read(sys.stdin.fileno(), 1)
            key_ord = ord(key)
            if is_input:
                keys.append(key)
            if not is_input and key_ord in _keys.BACKSPACE_OR_DEL:
                self.pending = self.pending[:-1]
                skip_process = True
                break
            else:
                self.pending.append(key_ord)
        if not skip_process:
            result, pending = self._process(self.pending)
            self.pending = pending
        else:
            result = []
        if is_input and keys:
            result.append(functools.partial(self._read_input_handler, keys))
        return result

    @staticmethod
    async def _read_input_handler(keys, viewer):
        if not viewer.is_input_mode:
            return
        writer = viewer.focused.holder.iterator.metadata.get("input")
        if not writer:
            return
        writer.write(b"".join(keys))
        await writer.drain()

    def _process(self, keys):
        viewer = self.viewer
        bindings = self.bindings
        if viewer.help.show:
            mode = _keys.HELP
        elif viewer.is_input_mode:
            mode = _keys.INPUT
        elif viewer.is_scrolling:
            mode = _keys.SCROLL
        else:
            mode = _keys.NORMAL
        sequences = [bindings[mode]]
        if mode != _keys.INPUT:
            sequences.append(bindings[_keys.GLOBAL])

        result = []
        while keys:
            has_pending = False
            found_sequence = False
            for mode_sequences in sequences:
                for key_sequence, fn in mode_sequences.items():
                    common_prefix = os.path.commonprefix([tuple(keys), key_sequence])
                    if common_prefix:
                        if len(common_prefix) == len(key_sequence):
                            result.append(fn)
                            keys = keys[len(key_sequence) :]
                            found_sequence = True
                            break
                        elif list(key_sequence)[: len(keys)] == keys:
                            has_pending = True
            if has_pending:
                break
            if not found_sequence:
                keys = keys[1:]
        return result, keys



from dataclasses import dataclass
from typing import List

from multiplex.refs import SPLIT


class Action:
    pass


class BoxAction:
    def run(self, box_holder):
        raise NotImplementedError


@dataclass
class SetTitle(BoxAction):
    title: str

    def run(self, box_holder):
        title = self.title
        iterator = box_holder.iterator
        if iterator.iterator is SPLIT:
            title += f" ({iterator.title})"
        iterator.title = title


class ToggleCollapse(BoxAction):
    def __init__(self, value=None):
        self.value = value

    def run(self, box_holder):
        box_holder.box.toggle_collapse(self.value)


class ToggleWrap(BoxAction):
    def __init__(self, value=None):
        self.value = value

    def run(self, box_holder):
        box_holder.box.toggle_wrap(self.value)


@dataclass
class UpdateMetadata(BoxAction):
    metadata: dict

    def run(self, box_holder):
        box_holder.iterator.metadata.update(self.metadata)


@dataclass
class BoxActions(BoxAction):
    actions: List[BoxAction]

    def run(self, box_holder):
        for action in self.actions:
            if isinstance(action, BoxAction):
                action.run(box_holder)
            elif callable(action):
                action(box_holder)
            else:
                raise RuntimeError(f"Invalid action: {action}")



import io

from easyansi import screen, cursor, drawing, attributes
from easyansi import colors_rgb as colors
from easyansi._core.codes import CSI


FULL_REFRESH = object()

buffer = io.StringIO()


def prnt(text):
    buffer.write(str(text))


def flush():
    global buffer
    if not buffer.tell():
        return
    screen.prnt(buffer.getvalue())
    buffer = io.StringIO()


NONE = object()
RESET = colors.reset_code()

BLACK = (0, 0, 0)
RED = (255, 0, 0)
GREEN = (0, 255, 0)
GRAY = (51, 51, 51)
BLUE1 = (37, 94, 132)
GREEN2 = (184, 212, 67)
ORANGE = (240, 140, 52)


class Theme:
    X_COLOR = (RED, None)
    V_COLOR = (GREEN, None)
    TITLE_NORMAL = (GRAY, NONE)
    TITLE_FOCUS = (BLUE1, NONE)
    TITLE_STREAM_DONE = (GRAY, NONE)
    STATUS_NORMAL = (NONE, GRAY)
    STATUS_SCROLL = (NONE, BLUE1)
    STATUS_INPUT = (BLACK, ORANGE)
    STATUS_SAVE = (BLACK, GREEN2)
    HELP_TITLES = (GREEN2, NONE)


theme = Theme


CLEAR_LINE = CSI + "2K"
ENABLE_ALT_BUFFER = CSI + "?1049h"
DISABLE_ALT_BUFFER = CSI + "?1049l"


def color_code(pair):
    fg, bg = pair
    if not fg or fg is NONE:
        fg = (None, None, None)
    if not bg or bg is NONE:
        bg = (None, None, None)
    return colors.color_code(*fg, *bg)


def setup():
    screen.prnt(ENABLE_ALT_BUFFER)
    cursor.hide()


def restore():
    global buffer
    clear()
    buffer = io.StringIO()
    cursor.show()
    screen.clear()
    screen.prnt(DISABLE_ALT_BUFFER)


def get_size():
    return screen.get_size()


def clear():
    prnt(screen.clear_code())


def text_box(from_row, to_row, text):
    for line_num in range(to_row, from_row - 1, -1):
        prnt(screen.clear_line_code(line_num))
    prnt(text)


def title(row, text, cols, hline_color):
    prnt(screen.clear_line_code(row))
    prnt(color_code(hline_color))
    prnt(drawing.hline_code(1))
    prnt(RESET)
    prnt(text)
    prnt(RESET)
    offset = 1 + len(text)
    prnt(color_code(hline_color))
    prnt(drawing.hline_code(cols - offset))
    prnt(RESET)


def status_bar(row, text):
    prnt(screen.clear_line_code(row))
    prnt(text)
    prnt(RESET)


def move_cursor(col, row):
    prnt(cursor.locate_code(col, row))


def show_cursor():
    prnt(cursor.show_code())


def hide_cursor():
    prnt(cursor.hide_code())


def help_screen(current_line, lines, cols, descriptions):
    prnt(screen.clear_code())
    prnt(drawing.box_code(width=cols, height=lines))
    prnt(cursor.locate_code(2, 1))

    help_lines = []
    help_line = io.StringIO()

    def next_line():
        nonlocal help_line
        help_line.write(cursor.next_line_code())
        help_line.write(cursor.right_code(2))
        help_lines.append(help_line.getvalue())
        help_line = io.StringIO()

    for mode, mode_desciptions in descriptions.items():
        help_line.write(attributes.bright_code())
        help_line.write(color_code(theme.HELP_TITLES))
        help_line.write(mode.capitalize())
        help_line.write(RESET)
        next_line()
        for keys, description in mode_desciptions.items():
            keys_text = ", ".join(str(k) for k in keys)
            line_text = f"{keys_text:<30}{description}"
            help_line.write(line_text)
            next_line()
        next_line()

    help_lines = help_lines[current_line : current_line + lines - 2]
    prnt("".join(help_lines))


class C:
    def __init__(self, *args, color=None):
        color = color or (None, None)
        self.color = color
        self.fg, self.bg = color
        self.parts = []
        for part in args:
            self._add(part)

    def copy(self):
        parts = [p.copy() if isinstance(p, C) else p for p in self.parts]
        return self.__class__(*parts, color=self.color)

    def __len__(self):
        return sum(len(p) for p in self.parts)

    def __str__(self):
        return self.to_string()

    def to_string(self, no_style=False):
        fg, bg = self.fg, self.bg
        style = None
        if fg is NONE and bg is NONE:
            style = RESET
            fg, bg = None, None
        elif fg is NONE:
            fg = (255, 255, 255)
        elif bg is NONE:
            bg = (0, 0, 0)

        if fg or bg:
            style = color_code((fg, bg))
        result = io.StringIO()
        if not no_style and style:
            result.write(style)
        for part in self.parts:
            value = part.to_string(no_style) if isinstance(part, C) else str(part)
            result.write(value)
            if not no_style and isinstance(part, C) and style:
                result.write(style)
        return result.getvalue()

    def __add__(self, other):
        copy = self.copy()
        copy._add(other)
        return copy

    def __getitem__(self, item):
        assert isinstance(item, slice)
        assert not item.start
        assert not item.step
        size = item.stop
        result = self.copy()
        result._truncate(size)
        return result

    def _truncate(self, size):
        new_parts = []
        for part in self.parts:
            part_len = len(part)
            if part_len <= size:
                new_parts.append(part)
                size -= part_len
            else:
                new_parts.append(part[:size])
                break
        self.parts = new_parts

    def _add(self, part):
        if isinstance(part, C):
            part = part.copy()
            if not part.fg:
                part.fg = self.fg
            if not part.bg:
                part.bg = self.bg
        self.parts.append(part)

    def to_dict(self):
        return {
            "color": self.color,
            "parts": [p.to_dict() if isinstance(p, C) else p for p in self.parts],
        }

    @staticmethod
    def from_dict(dct):
        color = dct["color"]
        parts = [C.from_dict(p) if isinstance(p, dict) else p for p in dct["parts"]]
        return C(*parts, color=color)
